"""
A suite of unit tests for testing utilities the GDS CLI uses
"""

from copy import deepcopy

import pytest

import fprime_gds.common.gds_cli.filtering_utils as filtering_utils
import fprime_gds.common.gds_cli.misc_utils as misc_utils
import fprime_gds.common.gds_cli.test_api_utils as test_api_utils
from fprime_gds.common.data_types.pkt_data import PktData
from fprime_gds.common.data_types.sys_data import SysData
from fprime_gds.common.templates.data_template import DataTemplate
from fprime_gds.common.templates.pkt_template import PktTemplate
from fprime_gds.common.testing_fw import predicates

# ==============================================================================
# Test misc_utils.repeat_until_interrupt
# ==============================================================================


@pytest.mark.gds_cli
def test_repeat_exits_at_counter():
    """
    Test basic functionality works as intended, can pass updated arguments, and
    exits on KeyboardInterrupt
    """
    count = 0

    def increment_count():
        nonlocal count
        count += 1

    def increment_with_interrupt(counter, exit_num, external_func):
        if counter == exit_num:
            raise KeyboardInterrupt
        external_func()
        return counter + 1, exit_num, external_func

    misc_utils.repeat_until_interrupt(increment_with_interrupt, 0, 3, increment_count)

    assert count == 3


@pytest.mark.gds_cli
def test_repeat_with_void_function_valid():
    """
    Test repeating with a function that doesn't return anything still works
    """
    count = 0

    def exit_at_value(exit_num):
        nonlocal count
        count += 1
        if count == exit_num:
            raise KeyboardInterrupt

    misc_utils.repeat_until_interrupt(exit_at_value, 5)

    assert count == 5


@pytest.mark.gds_cli
def test_repeat_with_no_arg_void_function_valid():
    """
    Test repeating with a function that takes no arguments doesn't throw an
    exception
    """
    count = 0

    def exit_when_3():
        nonlocal count
        count += 1
        if count == 3:
            raise KeyboardInterrupt

    misc_utils.repeat_until_interrupt(exit_when_3)

    assert count == 3


@pytest.mark.gds_cli
def test_repeat_with_no_arg_returning_function_invalid():
    """
    Test repeating with a function that takes no arguments but still returns a
    value is invalid
    """
    count = 0

    def error_before_exit():
        nonlocal count
        count += 1
        if count == 2:
            raise KeyboardInterrupt
        return "this should cause an error"

    with pytest.raises(TypeError):
        misc_utils.repeat_until_interrupt(error_before_exit)


@pytest.mark.gds_cli
def test_repeat_with_function_returning_naked_string_problematic():
    """
    Test that returning a string without encapsulating it in a tuple will fail,
    since every character will be counted as a different argument
    """
    count = 0

    def error_before_exit():
        nonlocal count
        count += 1
        if count == 2:
            raise KeyboardInterrupt
        return "this will return 49 arguments, because characters"

    with pytest.raises(TypeError):
        misc_utils.repeat_until_interrupt(error_before_exit)


@pytest.mark.gds_cli
def test_repeat_with_function_returning_enclosed_string_valid():
    """
    Test that returning a string when encapsulating it in a tuple will count
    it as just 1 argument
    """
    count = 0

    def error_before_exit(my_string):
        nonlocal count
        count += 1
        if count == 2:
            raise KeyboardInterrupt
        return (my_string,)  # this is only 1 argument, since it's in a tuple

    misc_utils.repeat_until_interrupt(error_before_exit, "let's do this")


# ==============================================================================
# Test test_api_utils.get_item_list
# ==============================================================================

# Pytest fixtures work by reusing outer names, so disable this warning
# pylint: disable=redefined-outer-name
@pytest.fixture
def item_template_dictionary():
    # Can't instantiate raw data template, so instead instantiating PktTemp
    template_12 = PktTemplate(pkt_id=12, pkt_name="OneTwo", ch_temp_list=[])
    template_34 = PktTemplate(pkt_id=34, pkt_name="ThreeFour", ch_temp_list=[])
    template_56 = PktTemplate(pkt_id=56, pkt_name="FiveSix", ch_temp_list=[])
    template_78 = PktTemplate(pkt_id=78, pkt_name="SevenEight", ch_temp_list=[])
    template_sur = PktTemplate(pkt_id=0, pkt_name="SURPRISE!", ch_temp_list=[])

    return {
        "12": template_12,
        "56": template_56,
        "34": template_34,
        "78": template_78,
        "SURPRISE!": template_sur,
    }


def t2d(template: DataTemplate) -> SysData:
    """
    Convert a DataTemplate object to a SysData object
    """
    # Currently unable to instantiate SysData, so instantiating PktData instead
    return PktData([], None, template)


def assert_equal_template_dictionaries(dict1, dict2):
    assert dict1.keys() == dict2.keys()
    for key in dict1:
        assert dict1[key].get_id() == dict2[key].get_id()
        assert dict1[key].get_name() == dict2[key].get_name()


def assert_equal_data_lists(list1, list2):
    assert len(list1) == len(list2)
    for item1, item2 in zip(list1, list2):
        assert item1.get_id() == item2.get_id()


@pytest.mark.gds_cli
def test_valid_get_item_list(item_template_dictionary):
    original_dict = deepcopy(item_template_dictionary)
    item_list = test_api_utils.get_item_list(
        item_dictionary=item_template_dictionary,
        search_filter=predicates.always_true(),
        template_to_data=t2d,
    )

    # Test all items present AND in ID-sorted order
    assert_equal_data_lists(
        item_list,
        [
            t2d(item_template_dictionary["SURPRISE!"]),
            t2d(item_template_dictionary["12"]),
            t2d(item_template_dictionary["34"]),
            t2d(item_template_dictionary["56"]),
            t2d(item_template_dictionary["78"]),
        ],
    )
    # Check nothing changed in the original dictionary
    assert_equal_template_dictionaries(item_template_dictionary, original_dict)


@pytest.mark.gds_cli
def test_get_item_list_with_id_filter(item_template_dictionary):
    original_dict = deepcopy(item_template_dictionary)
    item_list = test_api_utils.get_item_list(
        item_dictionary=item_template_dictionary,
        search_filter=filtering_utils.get_full_filter_predicate([12, 34], [], None),
        template_to_data=t2d,
    )

    assert_equal_data_lists(
        item_list,
        [
            t2d(item_template_dictionary["12"]),
            t2d(item_template_dictionary["34"]),
        ],
    )
    # Check nothing changed in the original dictionary
    assert_equal_template_dictionaries(item_template_dictionary, original_dict)


@pytest.mark.gds_cli
def test_get_empty_item_list_with_unused_id(item_template_dictionary):
    original_dict = deepcopy(item_template_dictionary)
    item_list = test_api_utils.get_item_list(
        item_dictionary=item_template_dictionary,
        search_filter=filtering_utils.get_full_filter_predicate([451], [], None),
        template_to_data=t2d,
    )

    assert item_list == []
    # Check nothing changed in the original dictionary
    assert_equal_template_dictionaries(item_template_dictionary, original_dict)


@pytest.mark.gds_cli
def test_get_item_list_with_empty_item_dictionary():
    item_list = test_api_utils.get_item_list(
        item_dictionary={}, search_filter=predicates.always_true(), template_to_data=t2d
    )

    assert item_list == []



import os
import sys
import unittest

from fprime_gds.common.history.test import TestHistory
from fprime_gds.common.testing_fw import predicates

filename = os.path.dirname(__file__)
gdsName = os.path.join(filename, "../../../../src")
fprimeName = os.path.join(filename, "../../../../../Fw/Python/src")
sys.path.insert(0, gdsName)
sys.path.insert(0, fprimeName)


class HistoryTestCases(unittest.TestCase):
    def setUp(self):
        self.tHistory = TestHistory()

    @staticmethod
    def assert_lists_equal(expected, actual):
        assert len(expected) == len(
            actual
        ), f"the given list should have had the length {len(expected)}, but instead had {len(actual)}"
        for i in range(len(expected)):
            assert (
                expected[i] == actual[i]
            ), f"the {i} element of the expected list should be {expected[i]}, but was {actual[i]}."

    def test_push_and_retrieve(self):
        tList = []
        for i in range(200):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.assert_lists_equal(tList, self.tHistory.retrieve())

    def test_push_and_retrieve_at_index(self):
        tList = []
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.tHistory.data_callback("hello")
        tList.append("hello")
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.assert_lists_equal(tList[10:], self.tHistory.retrieve(10))

    def test_push_and_retrieve_from_predicate(self):
        tList = []
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.tHistory.data_callback("hello")
        tList.append("hello")
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)
        start = predicates.equal_to("hello")
        self.assert_lists_equal(tList[50:], self.tHistory.retrieve(start))

    def test_push_and_clear(self):
        tList = []
        for i in range(200):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.assert_lists_equal(tList, self.tHistory.retrieve())
        self.tHistory.clear()
        self.assert_lists_equal([], self.tHistory.retrieve())

    def test_push_and_clear_index(self):
        tList = []
        for i in range(200):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.assert_lists_equal(tList, self.tHistory.retrieve())
        self.tHistory.clear(100)
        self.assert_lists_equal(tList[100:], self.tHistory.retrieve())

    def test_push_and_clear_predicate(self):
        tList = []
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.tHistory.data_callback("hello")
        tList.append("hello")
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)
        self.assert_lists_equal(tList, self.tHistory.retrieve())
        start = predicates.equal_to("hello")
        self.tHistory.clear(start)
        self.assert_lists_equal(tList[50:], self.tHistory.retrieve())

    def test_history_size(self):
        assert self.tHistory.size() == 0, "starting history is empty"
        for i in range(50):
            self.tHistory.data_callback(i)
        assert self.tHistory.size() == 50, "starting history is empty"
        self.tHistory.clear(25)
        assert self.tHistory.size() == 25, "starting history is empty"

    def test_history_retrieve_new(self):
        tList = []
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)

        self.assert_lists_equal(tList, self.tHistory.retrieve())
        tList.clear()

        self.tHistory.data_callback("hello")
        tList.append("hello")

        self.assert_lists_equal(tList, self.tHistory.retrieve_new())
        tList.clear()

        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)

        self.assert_lists_equal(tList, self.tHistory.retrieve_new())

    def test_history_retrieve_new_after_clear(self):
        tList = []
        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)

        self.assert_lists_equal(tList, self.tHistory.retrieve())
        tList.clear()

        for i in range(50):
            self.tHistory.data_callback(i)
            tList.append(i)

        self.tHistory.clear(75)

        self.assert_lists_equal(tList[25:], self.tHistory.retrieve_new())
        tList.clear()

    def test_history_iterable(self):
        for i in range(50):
            self.tHistory.data_callback(i)
        tList = []
        for item in self.tHistory:
            tList.append(item)
        self.assert_lists_equal(tList, self.tHistory.retrieve_new())

    def test_history_length(self):
        assert len(self.tHistory) == 0, "starting history is empty"
        for i in range(50):
            self.tHistory.data_callback(i)
        assert len(self.tHistory) == 50, "starting history is empty"
        self.tHistory.clear(25)
        assert len(self.tHistory) == 25, "starting history is empty"

    def test_history_filter(self):
        class is_even(predicates.predicate):
            def __call__(self, item):
                return item % 2 == 0

            def __str__(self):
                return "Decides if a value is even"

        pred = is_even()
        self.tHistory = TestHistory(filter_pred=pred)
        for i in range(50):
            self.tHistory.data_callback(i)
        assert len(self.tHistory) == 25
        self.assert_lists_equal(range(0, 50, 2), self.tHistory.retrieve())

        correct_error = False
        try:
            TestHistory("Not a predicate")
        except TypeError:
            correct_error = True
        assert correct_error, "The History should have raised a TypeError"


if __name__ == "__main__":
    unittest.main()



"""
uplinker.py:

Contains the code necessary to uplink files through the F prime ground system to a running deployment. The file uplink
process will read in a file from the OS, and uplink it in chunks. The system throttles chunks of the file requiring a
handshake packet in return before the next chunk is sent out.

@author lestarch
"""
import os
import queue
import threading
import time

import fprime_gds.common.handlers
from fprime_gds.common.data_types.file_data import (
    CancelPacketData,
    DataPacketData,
    EndPacketData,
    StartPacketData,
)
from fprime_gds.common.files.helpers import (
    FileStates,
    Timeout,
    TransmitFile,
    TransmitFileState,
    file_to_dict,
)


class UplinkQueue:
    """
    Handles queuing of files to send to the uplinker. This offloads the work of determining when done, and what to do
    while waiting. It also owns the thread that starts uplink. This thread watches for the current uplink to finish, and
    then it starts the next one and returns to a quiescent state.
    """

    def __init__(self, uplinker):
        """
        Constructs the uplink queue with a reference to the object that run the uplink.

        :param uplinker: uplinker to callback into
        """
        self.running = True
        self.uplinker = uplinker
        self.busy = threading.Semaphore()
        self.queue = queue.Queue()
        self.__file_store = []
        self.__exit = threading.Event()
        self.__thread = threading.Thread(target=self.run, name="UplinkerThread", args=())
        self.__thread.start()

    def enqueue(self, filepath, destination):
        """
        Enqueue the file and destination pair onto the queue

        :param filepath: filepath to upload to the given destination
        :param destination: destination path to upload the filepath to
        """
        file_obj = TransmitFile(filepath, destination)
        self.queue.put(file_obj)
        self.__file_store.append(file_obj)

    def pause(self):
        """Pause the uplinker, setting the running flag off, then holding the busy semaphore"""
        if self.running:
            self.running = False
            self.busy.acquire()

    def unpause(self):
        """Unpause the uplinker, releasing the busy semaphore and then restoring the running flag"""
        if not self.running:
            self.busy.release()
            self.running = True

    def is_running(self):
        """Check if the uplink is running"""
        return self.running

    def remove(self, source):
        """
        Remove a file by its source from the queue and the permanent uplink list. This will touch each record in the
        queue one time, as they are pulled off the queue, inspected, and the replaced.

        :param source: source file to remove
        """
        try:
            first = None
            found = self.queue.get_nowait()
            while found != first:
                if found.source == source:
                    break
                elif first is None:
                    first = found
            self.queue.put_nowait(found)
            found = self.queue.get_nowait()
        except queue.Empty:
            return
        self.__file_store.remove(found)

    def run(self):
        """
        A thread that will uplink files on after another until all files that have been enqueued are properly processed.
        To stop this thread, set the exit event.
        """
        while not self.__exit.is_set():
            # While running (not paused) continue to get - uplink - wait
            while self.running:
                file_obj = self.queue.get()
                # Exit condition is a None message, which unblocks the above get
                if file_obj is None or not self.running:
                    self.queue.put(file_obj)
                    break
                # Once a file is obtained from the queue, acquire the busy lock before starting. This prevents a in-use
                # or paused queue from uplinking.
                self.busy.acquire()
                self.uplinker.start(file_obj)
                # Wait until the file is finished, then release again waiting for more files
                self.busy.acquire()
                self.busy.release()
            time.sleep(0.010)  # Don't busy spin, but yield and sleep

    def current(self):
        """
        Gets a set of current files. This will create a copy, to prevent messing with uplink.

        :return: copy of current files transformed into a JSONable dictionary
        """
        return file_to_dict(self.__file_store)

    def exit(self):
        """Exit event to shutdown the thread"""
        self.__exit.set()
        self.running = False
        self.queue.put(
            None
        )  # Force an end to the wait for a file, if an uplink is not in-progress

    def join(self):
        """Join with this uplinker"""
        self.__thread.join()


class FileUplinker(fprime_gds.common.handlers.DataHandler):
    """
    File uplinking component. Keeps track of the currently uplinking file, and registers as a receiver for the handshake
    packets that are send back on each packet.
    """

    CHUNK_SIZE = 256

    def __init__(self, file_encoder, chunk=CHUNK_SIZE, timeout=20):
        """
        Constructor to build the file uplinker.
        """
        self.state = FileStates.IDLE
        self.queue = UplinkQueue(self)
        self.active = None
        self.sequence = 0
        self.chunk = chunk
        self.file_encoder = file_encoder
        self.__destination_dir = "/"
        self.__expected = []
        self.__timeout = Timeout()
        self.__timeout.setup(self.timeout, timeout)

    def enqueue(self, filepath, destination=None):
        """
        Enqueue files for the upload. This tunnels into the upload queue, which unblocks once files have been enqueued
        and begins to upload each file sequentially.

        :param filepath: filepath to upload to the system
        :param destination: (optional) destination to uplink to. Default: current destination + file's basename
        """
        if destination is None:
            destination = os.path.join(
                self.__destination_dir, os.path.basename(filepath)
            )
        self.queue.enqueue(filepath, destination)

    def exit(self):
        """Exit this uplinker by killing the thread"""
        self.queue.exit()
        self.cancel()
        self.queue.join()

    def is_running(self):
        """Check if the queue is running"""
        return self.queue.is_running()

    def pause(self):
        """Pause uplink by canceling the uplink, and then pausing the uplink queue"""
        self.cancel()
        self.queue.pause()

    def unpause(self):
        """Unpauses the uplink by unpausing the internal queue"""
        self.queue.unpause()

    def cancel_remove(self, file):
        """
        Cancel/remove the uplink of the given file.  If uplinking, it will cancel the uplink, and if queued it will
        remove the file from the queue.  Unknown files will be ignored.

        :param file: file to remove from the uplinker
        """
        if self.active is not None and file == self.active.source:
            self.cancel()
        else:
            self.queue.remove(file)

    def current_files(self):
        """
        Returns the current set of files held by the uplink queue.

        :return: current files as a list in JSONable format
        """
        return self.queue.current()

    def start(self, file_obj):
        """
        Starts a file uplink to a running F prime deployment. This will open of the file, save the file handle, and emit
        the start file uplink packet. It will also store the state for the file uplink progress. If already uplinking
        then a FileUplinkerBusyException will be raised.

        :param file_obj: file object to upload
        :raises FileUplinkerBusyException: when an upload is already running
        :raises OSError: when file is inaccessible
        """
        # Prevent multiple uplinks at once
        if self.state != FileStates.IDLE:
            raise FileUplinkerBusyException(
                f"Currently uplinking file '{self.active.source}' cannot start uplinking '{file_obj.source}'"
            )
        self.state = FileStates.RUNNING
        self.active = file_obj
        self.active.open(TransmitFileState.READ)
        self.send(
            StartPacketData(
                self.get_next_sequence(),
                self.active.size,
                self.active.source,
                self.active.destination,
            )
        )

    def send(self, packet_data):
        """
        A function to send the packet out.  Starts timeout and then pushes the packet to the file encoder.

        :param packet_data: packet data to send that will be pushed to the encoder
        """
        self.__timeout.restart()
        self.__expected = self.file_encoder.data_callback(packet_data)[8:]

    def data_callback(self, data, sender=None):
        """
        Process incoming handshake data, and if it is the expected handshake, then it uplinks the next packet. In the
        file. Invalid handshakes are ignore. When finished a returning handshake puts the uplinker into idle state.

        :param data: data from handshake packet to be verified against that previously sent
        """
        # Ignore handshakes not for us
        if not self.valid_handshake(data):
            return
        # If it is an end-wait or a cancel state, respond without reading next chunk
        if self.state == FileStates.END_WAIT:
            self.active.state = (
                "FINISHED" if self.active.state != "CANCELED" else "CANCELED"
            )
            self.state = FileStates.IDLE
            self.queue.busy.release()  # Allow the queue to continue
            self.__timeout.stop()
            return
        elif self.state == FileStates.CANCELED:
            self.send(CancelPacketData(self.get_next_sequence()))
            self.finish()
            return
        # Read next chunk of data.  b'' means the file is empty
        outgoing = self.active.read(self.chunk)
        if outgoing == b"":
            self.send(
                EndPacketData(self.get_next_sequence(), self.active.checksum.value)
            )
            self.finish()
        else:
            self.active.checksum.update(outgoing, self.active.seek)
            self.send(
                DataPacketData(self.get_next_sequence(), self.active.seek, outgoing)
            )
            self.active.seek += len(outgoing)

    def cancel(self):
        """
        Cancels the currently active file uplink. Will emit a cancel packet to inform the deployment that the file is
        canceled. This merely sets the state to "canceled" and will handle this later. This implies that in the case of
        a timeout, cancels may be ignored.
        """
        if self.state == FileStates.RUNNING:
            self.state = FileStates.CANCELED
            self.active.state = "CANCELED"
            # self.queue.pause()

    def timeout(self):
        """Handles timeout o file packet by finishing the upload immediately, and setting the state to timeout"""
        self.finish(False)
        self.active.state = "TIMEOUT"

    def finish(self, wait_for_handshake=True):
        """
        Finishes the current file uplink by closing the file, and starting the end process. If the uplinker should
        immediately terminate (like during a timeout) then set wait_for_handshake to False.

        :param wait_for_handshake: (optional) wrap up cleanly by waiting for handshake. Default: True, clean wait
        """
        self.active.close()
        self.sequence = 0
        self.state = FileStates.END_WAIT
        os.remove(self.active.source)
        # Immediate termination items
        if not wait_for_handshake:
            self.state = FileStates.IDLE
            self.__expected = []
            self.queue.busy.release()
            self.__timeout.stop()

    def get_next_sequence(self):
        """Gets the next sequence number"""
        tmp = self.sequence
        self.sequence = self.sequence + 1
        return tmp

    def valid_handshake(self, data):
        """
        Check the handshake data and ensure that it is as expected. This will allow us to only handle handshakes that
        we expected. This will ensure that the handshake data is an exact match of the send data.

        :param data: data to check against what was transmitted
        :return: True, if proper handshake, False otherwise
        """
        return data == self.__expected

    @property
    def destination_dir(self):
        """
        Get the destination directory that files will be uplinked to, if not currently specified.

        :return: value of destination
        """
        return self.__destination_dir

    @destination_dir.setter
    def destination_dir(self, destination):
        """
        Set the destination directory that files will be uplinked to.

        :param destination: new destination
        """
        self.__destination_dir = destination


class FileUplinkerBusyException(Exception):
    """File uplinker is busy and cannot uplink more files"""



"""
The implementation code for the command-send GDS CLI commands
"""

import difflib
from typing import Iterable, List

import fprime_gds.common.gds_cli.misc_utils as misc_utils
import fprime_gds.common.gds_cli.test_api_utils as test_api_utils
from fprime.common.models.serialize.type_exceptions import NotInitializedException
from fprime_gds.common.data_types.cmd_data import CommandArgumentsException
from fprime_gds.common.gds_cli.base_commands import QueryHistoryCommand
from fprime_gds.common.pipeline.dictionaries import Dictionaries
from fprime_gds.common.templates.cmd_template import CmdTemplate
from fprime_gds.common.testing_fw import predicates


class CommandSendCommand(QueryHistoryCommand):
    """
    The implementation for sending a command via the GDS to the spacecraft
    """

    @staticmethod
    def get_closest_commands(
        project_dictionary: Dictionaries, command_name: str, num: int = 3
    ) -> List[str]:
        """
        Searches for the closest matching known command(s) to the given command
        name.

        :param project_dictionary: The dictionary object for this project
            containing the item type definitions
        :param command_name: The full string name of the command to search for
        :param num: The maximum number of near-matches to return
        :return: A list of the closest matching commands (potentially empty)
        """
        known_commands = project_dictionary.command_name.keys()
        closest_matches = difflib.get_close_matches(command_name, known_commands, n=num)
        return closest_matches

    @staticmethod
    def get_command_template(
        project_dictionary: Dictionaries, command_name: str
    ) -> CmdTemplate:
        """
        Retrieves the command template for the given command name

        :param project_dictionary: The dictionary object for this project
            containing the item type definitions
        :param command_name: The full string name of the command to return a
            template for
        :return: The CmdTemplate object for the given command
        """
        return project_dictionary.command_name[command_name]

    @staticmethod
    def get_command_help_message(
        project_dictionary: Dictionaries, command_name: str
    ) -> str:
        """
        Returns a string showing a help message for the given GDS command.

        :param project_dictionary: The dictionary object for this project
            containing the item type definitions
        :param command_name: The full string name of the command to return a
            help message for
        :return: A help string for the command
        """
        command_template = CommandSendCommand.get_command_template(
            project_dictionary, command_name
        )
        return misc_utils.get_cmd_template_string(command_template)

    @classmethod
    def _get_item_list(
        cls,
        project_dictionary: Dictionaries,
        filter_predicate: predicates.predicate,
    ) -> Iterable[CmdTemplate]:
        """
        Gets a list of available commands in the system and return them in an
        ID-sorted list.

        :param project_dictionary: The dictionary object for the project
            containing the command definitions
        :param filter_predicate: Test API predicate used to filter shown
            channels
        """
        # NOTE: Trying to create a blank CmdData causes errors, so currently
        # just using templates (i.e. this function does nothing)
        def create_empty_command(cmd_template):
            return cmd_template

        command_list = test_api_utils.get_item_list(
            item_dictionary=project_dictionary.command_id,
            search_filter=filter_predicate,
            template_to_data=create_empty_command,
        )
        return command_list

    @classmethod
    def _get_upcoming_item(
        cls,
        api,
        filter_predicate,
        min_start_time="NOW",
        timeout=5.0,
    ):
        """
        NOTE: Doesn't use _get_upcoming_item; sign that this should not use QueryHistory as a base class,
        and should refactor when time's available
        """

    @classmethod
    def _get_item_string(
        cls,
        item: CmdTemplate,
        json: bool = False,
    ) -> str:
        """
        Converts the given command template into a human-readable string.

        :param item: The CmdTemplate to convert to a string
        :param json: Whether or not to return a JSON representation of "temp"
        :return: A readable string version of "item"
        """
        return misc_utils.get_cmd_template_string(item, json)

    @classmethod
    def _execute_command(
        cls,
        connection_info: misc_utils.ConnectionInfo,
        search_info: misc_utils.SearchInfo,
        command_name: str,
        arguments: List[str],
    ):
        """
        Handle the command-send arguments to connect to the Test API correctly,
        then send the command via the Test API.

        For more details on these arguments, see the command-send definition at:
            Gds/src/fprime_gds/executables/fprime_cli.py
        """
        dictionary, ip_address, port = tuple(connection_info)
        is_printing_list, ids, components, search, json = tuple(search_info)

        search_filter = cls._get_search_filter(ids, components, search, json)
        if is_printing_list:
            cls._log(cls._list_all_possible_items(dictionary, search_filter, json))
            return

        # ======================================================================
        pipeline, api = test_api_utils.initialize_test_api(
            dictionary, server_ip=ip_address, server_port=port
        )
        # ======================================================================

        try:
            api.send_command(command_name, arguments)
        except KeyError:
            cls._log(f"{command_name} is not a known command")
            close_matches = CommandSendCommand.get_closest_commands(
                pipeline.dictionaries, command_name
            )
            if close_matches:
                cls._log(f"Similar known commands: {close_matches}")
        except NotInitializedException:
            temp = CommandSendCommand.get_command_template(
                pipeline.dictionaries, command_name
            )
            cls._log(
                "'%s' requires %d arguments (%d given)"
                % (command_name, len(temp.get_args()), len(arguments))
            )
            cls._log(cls.get_command_help_message(pipeline.dictionaries, command_name))
        except CommandArgumentsException as err:
            cls._log(f"Invalid arguments given; {str(err)}")
            cls._log(cls.get_command_help_message(pipeline.dictionaries, command_name))

        # ======================================================================
        pipeline.disconnect()
        api.teardown()
        # ======================================================================

    @classmethod
    def handle_arguments(cls, *args, **kwargs):
        """
        Handle the given input arguments, then execute the command itself

        NOTE: This is currently just a pass-through method
        """
        connection_info = misc_utils.ConnectionInfo(
            kwargs["dictionary"], kwargs["ip_address"], kwargs["port"]
        )
        search_info = misc_utils.SearchInfo(
            kwargs["is_printing_list"],
            kwargs["ids"],
            kwargs["components"],
            kwargs["search"],
            kwargs["json"],
        )

        cls._execute_command(
            connection_info, search_info, kwargs["command_name"], kwargs["arguments"]
        )



"""
string_util.py
Utility functions to process strings to be used in FPrime GDS
@Created March 18, 2021
@`janamian`

Note: This function has an identical copy in fprime-gds
"""

import re
import logging

LOGGER = logging.getLogger("string_util_logger")


def format_string_template(format_str, given_values):
    r"""
    Function to convert C-string style to python format
    without using python interpolation
    Considered the following format for C-string:
    %[flags][width][.precision][length]type

    0- %:                (?<!%)(?:%%)*%
    1- flags:            ([\-\+0\ \#])?
    2- width:            (\d+|\*)?
    3- .precision:       (\.\*|\.\d+)?
    4- length:          `([hLIw]|l{1,2}|I32|I64)?`
    5- conversion_type: `([cCdiouxXeEfgGaAnpsSZ])`

    Note:
    This function will keep the flags, width, and .precision of C-string
    template.

    It will keep f, d, x, o, and e flags and remove all other types.
    Other types will be duck-typed by python
    interpreter.

    lengths will also be removed since they are not meaningful to Python interpreter.
    `See: https://docs.python.org/3/library/stdtypes.html#printf-style-string-formatting`

    `Regex Source: https://www.regexlib.com/REDetails.aspx?regexp_id=3363`
    """

    def convert(match_obj, ignore_int):
        if match_obj.group() is not None:
            flags, width, precision, length, conversion_type = match_obj.groups()
            format_template = ""
            if flags:
                format_template += f"{flags}"
            if width:
                format_template += f"{width}"
            if precision:
                format_template += f"{precision}"

            if conversion_type:
                if any(
                    [
                        str(conversion_type).lower() == "f",
                        str(conversion_type).lower() == "x",
                        str(conversion_type).lower() == "o",
                        str(conversion_type).lower() == "e",
                    ]
                ):
                    format_template += f"{conversion_type}"
                elif all([not ignore_int, str(conversion_type).lower() == "d"]):
                    format_template += f"{conversion_type}"

            if format_template == "":
                template = "{}"
            else:
                template = "{:" + format_template + "}"
            return template
        return match_obj

    def convert_include_all(match_obj):
        return convert(match_obj, ignore_int=False)

    def convert_ignore_int(match_obj):
        return convert(match_obj, ignore_int=True)

    # Allowing single, list and tuple inputs
    if not isinstance(given_values, (list, tuple)):
        values = (given_values,)
    elif isinstance(given_values, list):
        values = tuple(given_values)
    else:
        values = given_values

    pattern = r"(?<!%)(?:%%)*%([\-\+0\ \#])?(\d+|\*)?(\.\*|\.\d+)?([hLIw]|l{1,2}|I32|I64)?([cCdiouxXeEfgGaAnpsSZ])"

    match = re.compile(pattern)

    # First try to include all types
    try:
        formatted_str = re.sub(match, convert_include_all, format_str)
        result = formatted_str.format(*values)
        result = result.replace("%%", "%")
        return result
    except Exception as exc:
        msg = "Value and format string do not match. "
        msg += " Will ignore integer flags `d` in string template. "
        msg += f"values: {values}. "
        msg += f"format_str: {format_str}. "
        msg += f"given_values: {given_values}"
        LOGGER.warning(msg)

    # Second try by not including %d.
    # This will resolve failing ENUMs with %d
    # but will fail on other types.
    try:
        formatted_str = re.sub(match, convert_ignore_int, format_str)
        result = formatted_str.format(*values)
        result = result.replace("%%", "%")
        return result
    except ValueError as e:
        msg = "Value and format string do not match. "
        msg += f"values: {values}. "
        msg += f"format_str: {format_str}. "
        msg += f"given_values: {given_values}"
        msg += f"Err Msg: {str(e)}\n"
        LOGGER.error(msg)
        raise ValueError



"""
Tests the event encoder

Created on Jul 10, 2020
@author: Joseph Paetz, hpaulson
"""


from fprime_gds.common.encoders.event_encoder import EventEncoder
from fprime_gds.common.data_types.event_data import EventData
from fprime_gds.common.utils.config_manager import ConfigManager
from fprime_gds.common.templates.event_template import EventTemplate
from fprime.common.models.serialize.time_type import TimeType
from fprime.common.models.serialize.numerical_types import U8Type, U16Type, U32Type
from fprime_gds.common.utils.event_severity import EventSeverity


def test_event_encoder():
    """
    Tests the encoding of the event encoder
    """
    config = ConfigManager()
    config.set("types", "msg_len", "U16")

    enc = EventEncoder()
    enc_config = EventEncoder(config)

    temp = EventTemplate(
        101,
        "test_ch",
        "test_comp",
        [("a1", "a1", U32Type), ("a2", "a2", U32Type)],
        EventSeverity["DIAGNOSTIC"],
        "%d %d",
    )

    time_obj = TimeType(2, 0, 1533758629, 123456)

    event_obj = EventData((U32Type(42), U32Type(10)), time_obj, temp)

    desc_bin = b"\x00\x00\x00\x02"
    id_bin = b"\x00\x00\x00\x65"
    time_bin = b"\x00\x02\x00\x5b\x6b\x4c\xa5\x00\x01\xe2\x40"
    arg_bin = b"\x00\x00\x00\x2a\x00\x00\x00\x0a"
    long_len_bin = b"\x00\x00\x00\x1b"
    short_len_bin = b"\x00\x1b"

    reg_expected = long_len_bin + desc_bin + id_bin + time_bin + arg_bin
    config_expected = short_len_bin + desc_bin + id_bin + time_bin + arg_bin

    reg_output = enc.encode_api(event_obj)

    assert (
        reg_output == reg_expected
    ), f"FAIL: expected regular output to be {list(reg_expected)}, but found {list(reg_output)}"

    config_output = enc_config.encode_api(event_obj)

    assert (
        config_output == config_expected
    ), f"FAIL: expected configured output to be {list(config_expected)}, but found {list(config_output)}"

    temp = EventTemplate(
        102,
        "test_ch2",
        "test_comp2",
        [("a1", "a1", U8Type), ("a2", "a2", U16Type)],
        EventSeverity["DIAGNOSTIC"],
        "%d %d",
    )

    time_obj = TimeType(2, 0, 1533758628, 123457)

    event_obj = EventData((U8Type(128), U16Type(40)), time_obj, temp)

    desc_bin = b"\x00\x00\x00\x02"
    id_bin = b"\x00\x00\x00\x66"
    time_bin = b"\x00\x02\x00\x5b\x6b\x4c\xa4\x00\x01\xe2\x41"
    arg_bin = b"\x80\x00\x28"
    long_len_bin = b"\x00\x00\x00\x16"
    short_len_bin = b"\x00\x16"

    reg_expected = long_len_bin + desc_bin + id_bin + time_bin + arg_bin
    config_expected = short_len_bin + desc_bin + id_bin + time_bin + arg_bin

    reg_output = enc.encode_api(event_obj)

    assert (
        reg_output == reg_expected
    ), f"FAIL: expected regular output to be {list(reg_expected)}, but found {list(reg_output)}"

    config_output = enc_config.encode_api(event_obj)

    assert (
        config_output == config_expected
    ), f"FAIL: expected configured output to be {list(config_expected)}, but found {list(config_output)}"



"""
A set of utility classes and functions for filtering the items we want to return
to the user in the GDS CLI
"""

from typing import Any, Callable, Iterable

from fprime_gds.common.data_types.cmd_data import CmdData
from fprime_gds.common.data_types.sys_data import SysData
from fprime_gds.common.testing_fw import predicates


class id_predicate(predicates.predicate):
    def __init__(self, id_num: int):
        """
        A predicate that tests if the SysData or DataTemplate argument given to
        it is of a given ID type.

        :param id_num: The ID to compare the item against
        """
        self.id_num = id_num

    def __call__(self, item):
        """

        :param item: The object or value to evaluate
        """
        return hasattr(item, "get_id") and self.id_num == item.get_id()

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x.id == {self.id_num}"


def get_id_predicate(ids: Iterable[int]) -> predicates.predicate:
    """
    Returns a Test API predicate that only accepts items with one of the given
    type IDs (if no IDs are given, accept all items).

    :param ids: A list of possible ID values to accept (empty if we should
        accept SysData with any ID)
    :return: A predicate that checks if a SysData object has one of the given
        IDs
    """
    if ids:
        id_preds = [id_predicate(id_num) for id_num in ids]
        return predicates.satisfies_any(id_preds)
    return predicates.always_true()


class component_predicate(predicates.predicate):
    def __init__(self, component: str):
        """
        A predicate that tests if the SysData or DataTemplate argument given is
        from the given component. If there is no component information found,
        returns True.

        :param component: The component name to check for
        """
        self.comp = component

    def __call__(self, item):
        """

        :param item: the object or value to evaluate
        """
        # NOTE: Always returns true if no component found
        item_component = self.comp
        if hasattr(item, "get_comp_name"):
            item_component = item.get_comp_name()
        elif hasattr(item.get_template(), "get_comp_name"):
            # Technically law of Minerva violation (acceptable?)
            item_component = item.get_template().get_comp_name()
        return self.comp == item_component

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f'x is in component "{self.comp}"'


def get_component_predicate(components: Iterable[str]) -> predicates.predicate:
    """
    Returns a Test API predicate that only accepts items from one of the given
    components (if no components are given, accept all items).

    :param components: A list of possible components to accept (empty if we
        should accept SysData from any component)
    :return: A predicate that checks if a SysData object come from one of the
        given components
    """
    if components:
        component_preds = [component_predicate(comp) for comp in components]
        return predicates.satisfies_any(component_preds)
    return predicates.always_true()


class contains_search_string(predicates.predicate):
    def __init__(self, search_string: str, to_string_func: Callable[[Any], str] = str):
        """
        A predicate that tests if the argument given to it contains the
        passed-in string (after the argument is converted to a string via
        __str__).

        :param search_string: The exact text to check for inside the object
        :param to_string_func: An optional method for converting the given
            object to a string
        """
        self.search_string = str(search_string)
        self.to_str = to_string_func

    def __call__(self, item):
        """

        :param item: the object or value to evaluate
        """
        return self.search_string in self.to_str(item)

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f'str(x) contains "{self.search_string}"'


def get_search_predicate(
    search_string: str, to_str: Callable[[Any], str] = str
) -> predicates.predicate:
    """
    Returns a Test API predicate that only accepts items whose string
    representation contains the exact given term.

    :param search_string: The substring to look for in the given object
    :param to_str: An optional function for converting a given object to a
        string
    :return: A predicate that checks if an object contains "search_string" when
        it's converted to a string
    """
    if search_string:
        return contains_search_string(search_string, to_str)
    return predicates.always_true()


def get_full_filter_predicate(
    ids: Iterable[int],
    components: Iterable[str],
    search_string: str,
    to_str: Callable[[Any], str] = str,
) -> predicates.predicate:
    """
    Returns a Test API predicate to only get recent data from the specified
    IDs/components, and containing the given search string. If any of these
    are left blank, no restrictions are assumed for that field.

    :param ids: A list of possible ID values to accept (empty if we should
        accept SysData with any ID)
    :param components: A list of possible components to accept (empty if we
        should accept SysData from any component)
    :param search_string: The substring to look for in the given object
    :param to_str: An optional function for converting a given object to a
        string

    :return: A predicate that only accepts SysData objects with the given
        IDs, components, and search term
    """
    return_all = predicates.always_true()

    id_pred = get_id_predicate(ids)
    comp_pred = get_component_predicate(components)
    search_pred = get_search_predicate(search_string, to_str)

    return predicates.satisfies_all([return_all, id_pred, comp_pred, search_pred])


class cmd_predicate(predicates.predicate):
    def __init__(self):
        """
        A predicate for specifying a CmdData object from data_types.cmd_data.
        This predicate can be used to search a history.
        """

    def __call__(self, cmd):
        """
        The cmd_predicate checks that the object is an instance of CmdData.

        :param cmd: an object to check for being a CmdData object
        """
        return isinstance(cmd, CmdData)

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return "True IFF: x is a CmdData object"


class time_to_data_predicate(predicates.predicate):
    def __init__(self, time_predicate: predicates.predicate):
        """
        Converts the given predicate from one called on an F' TimeType to one
        that can be called on a SysData type
        :param time_predicate: a predicate that expects a TimeType object to be
            passed in
        """
        self.time_pred = time_predicate

    def __call__(self, item: SysData):
        """
        Gets the given item's time and checks it using the passed-in predicate

        :param item: an object to check for being a CmdData object
        """
        return self.time_pred(item.get_time())

    def __str__(self):
        """
        Returns a string outlining the evaluation done by the predicate.
        """
        return f"x = x.get_time(), {self.time_pred}"



"""
@brief Loader class for importing python based event dictionaries

@date Created July 3, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime.common.models.serialize.type_exceptions import TypeMismatchException
from fprime_gds.common.templates import event_template
from fprime_gds.common.utils.event_severity import EventSeverity

# Custom Python Modules
from . import python_loader


class EventPyLoader(python_loader.PythonLoader):
    """Class to load python based event dictionaries"""

    # Field names in the python module file dictionaries
    ID_FIELD = "ID"
    NAME_FIELD = "NAME"
    COMP_FIELD = "COMPONENT"
    SEVERITY_FIELD = "SEVERITY"
    FMT_STR_FIELD = "FORMAT_STRING"
    DESC_FIELD = "EVENT_DESCRIPTION"
    ARGS_FIELD = "ARGUMENTS"

    def construct_dicts(self, path):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead use
        get_id_dict(path) and get_name_dict(path)

        Args:
            path: Path to the python module file dictionary to convert. This
                  should be a directory. If using a regular fprime deployment,
                  this should be a path to the events dictionary in your
                  generated folder:
                  ${GENERATED_FOLDER_LOCATION}/generated/${DEPLOYMENT}/events

        Returns:
            A tuple with two event dictionaries (python type dict):
            (id_dict, name_dict). The keys should be the events' id and name
            fields respectively and the values should be EventTemplate
            objects.
        """
        module_dicts = self.read_dict(path, use_superpkg=True)

        id_dict = {}
        name_dict = {}

        for event_dict in module_dicts:

            try:
                # Create an event template object
                event_temp = event_template.EventTemplate(
                    event_dict[self.ID_FIELD],
                    event_dict[self.NAME_FIELD],
                    event_dict[self.COMP_FIELD],
                    event_dict[self.ARGS_FIELD],
                    EventSeverity[event_dict[self.SEVERITY_FIELD]],
                    event_dict[self.FMT_STR_FIELD],
                    event_dict[self.DESC_FIELD],
                )
                id_dict[event_dict[self.ID_FIELD]] = event_temp
                name_dict[event_dict[self.NAME_FIELD]] = event_temp
            except TypeMismatchException as error:
                print(f"Type mismatch: {error.getMsg()}")
                raise error

        return id_dict, name_dict



""" Uplink and Downlink handling for communications layer

Downlink needs to happen in several stages. First, raw data is read from the adapter. This data is collected in a pool
and the pool is passed to a deframer that extracts frames from this pool. Frames are queued and sent to the ground
side where they are and passed into the ground side handler and onto the other GDS processes. Downlink handles multiple
streams of data the FSW downlink, and loopback data from the uplink adapter.

Uplink is the reverse, it pulls data in from the ground handler, frames it, and sends it up to the waiting FSW. Uplink
is represented by a single thread, as it is not dealing with multiple streams of data that need to be multiplexed.

"""
import threading
from queue import Queue, Full, Empty
import logging

from fprime.common.models.serialize.numerical_types import U32Type
from fprime_gds.common.utils.data_desc_type import DataDescType
from fprime_gds.common.communication.adapters.base import BaseAdapter
from fprime_gds.common.communication.ground import GroundHandler
from fprime_gds.common.communication.framing import FramerDeframer


DW_LOGGER = logging.getLogger("downlink")
UP_LOGGER = logging.getLogger("uplink")


class Downlinker:
    """Encapsulates communication downlink functions

    Handles downlink creating two threads, one to read and deframe, and the other to send data out to the ground side
    of the system. It is composed of an adapter used to read from the interface, a deframer that is used to deframe
    incoming data, and a ground handler that is used to interact with the ground side of the system.

    Two threaded stages are used to multiplex between loopback data and FSW downlink data without the need to busy spin
    waiting for data.
    """

    def __init__(
        self, adapter: BaseAdapter, ground: GroundHandler, deframer: FramerDeframer
    ):
        """Initialize the downlinker

        Constructs a new downlinker object used to run the downlink and deframing operation.

        Args:
            adapter: adapter used to read raw data from the hardware connection
            ground: handles the ground side connection
            deframer: deframer used to deframe data from the communication format
        """
        self.running = True
        self.th_ground = None
        self.th_data = None
        self.adapter = adapter
        self.ground = ground
        self.deframer = deframer
        self.outgoing = Queue()

    def start(self):
        """Starts the downlink pipeline"""
        self.th_ground = threading.Thread(target=self.sending, name="DownlinkTTSGroundThread")
        self.th_ground.daemon = True
        self.th_ground.start()
        self.th_data = threading.Thread(target=self.deframing, name="DownLinkDeframingThread")
        self.th_data.daemon = True
        self.th_data.start()

    def deframing(self):
        """Deframing stage of downlink

        Reads in data from the raw adapter and runs the deframing. Collects data in a pool and continually runs
        deframing against it where possible. Then appends new frames into the outgoing queue.
        """
        pool = b""
        while self.running:
            # Blocks until data is available, but may still return b"" if timeout
            pool += self.adapter.read()
            frames, pool = self.deframer.deframe_all(pool, no_copy=True)
            try:
                for frame in frames:
                    self.outgoing.put_nowait(frame)
            except Full:
                DW_LOGGER.warning("GDS ground queue full, dropping frame")

    def sending(self):
        """Outgoing stage of downlink

        Adapts the downlink adapter to the rest of the GDS system by draining the outgoing queue and sending those
        packets to the rest of the GDS. This uses the ground send_all method.
        """
        while self.running:
            frames = []
            try:
                # Blocking read of at least one frame, then drain the entire queue
                frames.append(self.outgoing.get(timeout=0.500))
                while not self.outgoing.empty():
                    frames.append(self.outgoing.get_nowait())
            except Empty:
                pass
            self.ground.send_all(frames)

    def stop(self):
        """Stop the thread depends will close the ground resource which may be blocking"""
        self.running = False

    def join(self):
        """Join on the ending threads"""
        for thread in [self.th_data, self.th_ground]:
            if thread is not None:
                thread.join()

    def add_loopback_frame(self, frame):
        """Adds a frame to loopback to ground

        Some uplink processes are virtualized on the ground, and thus must loopback packets. This is used for data
        handshaking that the FSW may not support.

        Args:
            frame: frame to loopback to ground
        """
        try:
            self.outgoing.put_nowait(frame)
        except Full:
            DW_LOGGER.warning("GDS ground queue full, dropping loopback frame")


class Uplinker:
    """Uplinker used to pull data out of the ground layer and send to FSW

    Handles uplink by creating a single thread to read data from the ground layer, frame it, and pass it to the adapter
    to the hardware link to flight software. It is composed of an adapter used to write to the interface, a framer
    that is used to frame outgoing data, and a ground handler that is used to interact with the ground side of the
    system.

    Since there is one stream of data the uplink requires only one thread to run.

    """

    RETRY_COUNT = 3

    def __init__(
        self,
        adapter: BaseAdapter,
        ground: GroundHandler,
        framer: FramerDeframer,
        loopback: Downlinker,
    ):
        """Initializes the uplink class

        Initialize the uplink class using a hardware adapter, ground handler, and framer.
        loopback is used to virtualize the return packet handshake as FSW does not handle that.

        Args:
            adapter: hardware adapter used to write raw outgoing data bytes
            ground: ground handler receiving data from the ground system
            framer: framer used to frame wire bytes
            loopback: used to return handshake packets
        """
        self.th_uplink = None
        self.running = True
        self.ground = ground
        self.adapter = adapter
        self.loopback = loopback
        self.framer = framer

    def start(self):
        """Starts the uplink pipeline"""
        self.th_uplink = threading.Thread(target=self.uplink, name="UplinkThread")
        self.th_uplink.daemon = True
        self.th_uplink.start()

    def uplink(self):
        """Runs uplink of data from ground to FSW

        Primary stage of the uplink process, reads data from the ground adapter, and passes the rest of the data to the
        framer, and then onto the adapter to send to FSW. Uplink also generates handshake packets as the current FSW
        does not generate handshake packets.
        """
        try:
            while self.running:
                packets = self.ground.receive_all()
                for packet in [
                    packet
                    for packet in packets
                    if packet is not None and len(packet) > 0
                ]:
                    framed = self.framer.frame(packet)
                    # Uplink handles synchronous retries
                    for retry in range(0, Uplinker.RETRY_COUNT):
                        if self.adapter.write(framed):
                            self.loopback.add_loopback_frame(
                                Uplinker.get_handshake(packet)
                            )
                            break
                    else:
                        UP_LOGGER.warning(
                            "Uplink failed to send %d bytes of data after %d retries",
                            len(framed),
                            Uplinker.RETRY_COUNT,
                        )
        # An OSError might occur during shutdown and is harmless. If we are not shutting down, this error should be
        # propagated up the stack.
        except OSError:
            if self.running:
                raise

    def stop(self):
        """Stop the thread depends will close the ground resource which may be blocking"""
        self.running = False

    def join(self):
        """Join on the ending threads"""
        if self.th_uplink is not None:
            self.th_uplink.join()

    @staticmethod
    def get_handshake(packet: bytes) -> bytes:
        """Gets a handshake raw frame from the last packet

        Creates a handshake raw-frame by repeating the contents of the last packet with a handshake descriptor at the
        front.

        Args:
            packet: packet to repeat back out as handshake

        Returns:
            handshake packet
        """
        return U32Type(DataDescType["FW_PACKET_HAND"].value).serialize() + packet



"""
cli.py:

This file sets up the command line interface and argument parsing that is done to support the F prime executable tools
layer. It is designed to allow users to import standard sets of arguments that applied to the various aspects of the
code that they are importing.

@author mstarch
"""
import abc
import argparse
import copy
import datetime
import errno
import importlib
import os
import re
import sys

import fprime_gds.common.communication.adapters.base
import fprime_gds.common.communication.checksum
import fprime_gds.common.logger

# Include basic adapters
import fprime_gds.common.communication.adapters.ip
import fprime_gds.common.utils.config_manager

try:
    import fprime_gds.common.communication.adapters.uart
except ImportError:
    pass
try:
    import zmq
except ImportError:
    zmq = None
# Try to import each GUI type, and if it can be imported
# it will be provided to the user as an option
GUIS = ["none", "html"]
try:
    import fprime_gds.wxgui.tools.gds

    GUIS.append("wx")
except ImportError:
    pass


class ParserBase(abc.ABC):
    """
    Parsers must define several functions. They must define "get_parser", which will produce a parser to parse the
    arguments, and an optional "handle_arguments" function to do any necessary processing of the arguments. Note: when
    handling arguments, the implementer should copy the incoming namespace if the original values of arguments will be
    modified.
    """

    @staticmethod
    @abc.abstractmethod
    def get_parser():
        """
        Produce a parser that will handle the given arguments. These parsers can be combined for a CLI for a tool by
        assembling them as parent processors to a parser for the given tool.
        """

    @classmethod
    @abc.abstractmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Handle arguments from the given parser. The expectation is that the "args" namespace is taken in, processed, and
        a new namespace object is returned with the processed variants of the arguments. Copy the namespace object when
        modifying existing arguments.

        :param args: arguments namespace of processed arguments
        :return: namespace with processed results of arguments.
        """

    @staticmethod
    def parse_args(
        parser_classes, description="No tool description provided", **kwargs
    ):
        """
        Create a parser for the given application using the description provided. This will then add all specified
        ParserBase subclasses' get_parser output as parent parses for the created parser. Then all of the handle
        arguments methods will be called, and the final namespace will be returned.

        :param parser_classes: a list of ParserBase subclasses that will be used to
        :param description: description passed ot the argument parser
        :return: namespace with all parsed arguments from all provided ParserBase subclasses
        """
        if kwargs is None:
            kwargs = {}
        created_parsers = [parser_base.get_parser() for parser_base in parser_classes]
        parser = argparse.ArgumentParser(
            description=description, parents=created_parsers
        )
        args = parser.parse_args()
        # Handle all arguments
        try:
            for parser_base in parser_classes:
                args = parser_base.handle_arguments(args, **kwargs)
        except ValueError as ver:
            print(f"[ERROR] Failed to parse arguments: {ver}", file=sys.stderr)
            sys.exit(-1)
        return args, parser

    @staticmethod
    def find_in(token, deploy, is_file=True):
        """
        Find token in deploy directory by walking the directory looking for reg-ex. This effectively finds a file in a
        subtree and provides the path to it. Returns None when not found

        :param token: token to search for in the directory structure
        :param deploy: directory to start with
        :param is_file: true if looking for file, otherwise false
        :return: full path to token in tree
        """
        for dirpath, dirs, files in os.walk(deploy):
            for check in files if is_file else dirs:
                if re.match(f"^{str(token)}$", check):
                    return os.path.join(dirpath, check)
        return None


class ImportParser(ParserBase):
    """
    Parser used to import new modules into the F prime namespace. This is how F prime can handle new modules and plugins
    by allowing the user to import new Python modules in, which may define new adapters and other items.
    """

    @staticmethod
    def get_parser():
        """
        Creates a parser that reads '--import' arguments to import new modules.  Multiple '--import' are supported.

        :return: parser that can handle imports
        """
        parser = argparse.ArgumentParser(
            description="Process import arguments allowing for addition of modules",
            add_help=False,
        )
        parser.add_argument(
            "--include",
            action="append",
            default=[],
            help="Module name in PYTHONPATH to import. Users may specify multiple --include flags.",
        )
        return parser

    @classmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Handle the import arguments by looping through them and importing each. Errors are printed to the console if an
        import failed, but no other effect is realized.

        :param args: parsed arguments namespace
        :return: args as input, has side effect of new imported modules
        """
        for module in args.include:
            try:
                globals()[module] = importlib.import_module(module)
            except ImportError as ime:
                print(
                    f"[WARNING] Failed to import '{module}' with error {ime}",
                    file=sys.stderr,
                )
        return args

    @staticmethod
    def run_first():
        """
        Runs this parser once first so that other parsers can benefit from the existing imported libraries. If there is
        a problem, the help message will be minimal, but should work.
        """
        args, _ = ImportParser.get_parser().parse_known_args()
        ImportParser.handle_arguments(args)


class CommParser(ParserBase):
    """
    Handles parsing of all of the comm-layer arguments. This means selecting a comm adapter, and passing the arguments
    required to setup that comm adapter. In addition, this parser uses the import parser to import modules such that a
    user may import other adapter implementation files.
    """

    @staticmethod
    def get_parser():
        """
        Produce a parser that will handle the given arguments. These parsers can be combined for a CLI for a tool by
        assembling them as parent processors to a parser for the given tool.
        """
        ImportParser.run_first()
        adapters = (
            fprime_gds.common.communication.adapters.base.BaseAdapter.get_adapters().keys()
        )
        adapter_parents = []
        for adapter_name in adapters:
            adapter = fprime_gds.common.communication.adapters.base.BaseAdapter.get_adapters()[
                adapter_name
            ]
            # Check adapter real quick before moving on
            if not hasattr(adapter, "get_arguments") or not callable(
                getattr(adapter, "get_arguments", None)
            ):
                print(
                    f"[WARNING] '{adapter_name}' does not have 'get_arguments' method, skipping.",
                    file=sys.stderr,
                )
                continue
            comm_parser = argparse.ArgumentParser(
                description=f"'{adapter_name}' parser", add_help=False
            )
            # Add arguments for the parser
            for argument in adapter.get_arguments().keys():
                comm_parser.add_argument(*argument, **adapter.get_arguments()[argument])
            adapter_parents.append(comm_parser)
        parser = argparse.ArgumentParser(
            description="Process arguments needed to specify a comm-adapter",
            add_help=False,
            parents=adapter_parents,
        )
        parser.add_argument(
            "--comm-adapter",
            dest="adapter",
            action="store",
            type=str,
            help="Adapter for communicating to flight deployment. [default: %(default)s]",
            choices=["none"] + list(adapters),
            default="ip",
        )
        parser.add_argument(
            "--comm-checksum-type",
            dest="checksum_type",
            action="store",
            type=str,
            help="Setup the checksum algorithm. [default: %(default)s]",
            choices=[
                item
                for item in fprime_gds.common.communication.checksum.CHECKSUM_MAPPING.keys()
                if item != "default"
            ],
            default=fprime_gds.common.communication.checksum.CHECKSUM_SELECTION,
        )
        return parser

    @classmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Handle the input arguments for the the parser. This will help setup the adapter with its expected arguments.

        :param args: parsed arguments in namespace format
        :return: namespace with "comm_adapter" value added
        """
        args = copy.copy(args)
        args.comm_adapter = (
            fprime_gds.common.communication.adapters.base.BaseAdapter.construct_adapter(
                args.adapter, args
            )
        )
        fprime_gds.common.communication.checksum.CHECKSUM_SELECTION = args.checksum_type
        return args


class LogDeployParser(ParserBase):
    """
    A parser that handles log files by reading in a '--logs' directory or a '--deploy' directory to put the logs into
    as a default. This is useful as a parsing fragment for any application that produces log files and needs these logs
    to end up in the proper place.
    """

    # Class variable to store logs on first encounter of the logging call. This helps ensure that the logs all end up in
    # a consistent directory, opposed to multiple directories across the system.
    first_log = None

    @staticmethod
    def get_parser():
        """
        Creates a new parser that can process "--deployment" and "--logs" arguments for use with any CLI producing log
        files. Can be used to construct parent arguments

        :return: parser with logging and deploy arguments
        """
        parser = argparse.ArgumentParser(
            description="Process arguments needed to specify a logging", add_help=False
        )
        parser.add_argument(
            "-l",
            "--logs",
            dest="logs",
            action="store",
            default=os.path.join(os.getcwd(), "logs"),
            type=str,
            help="Logging directory. Created if non-existent. [default: %(default)s]",
        )
        parser.add_argument(
            "--log-directly",
            dest="log_directly",
            action="store_true",
            default=False,
            help="Logging directory is used directly, no extra dated directories created.",
        )
        parser.add_argument("--log-to-stdout", action="store_true", default=False,
                            help="Log to standard out along with log output files")
        return parser

    @classmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Read the arguments specified in this parser and validate the expected inputs.

        :param args: parsed arguments as namespace
        :return: args namespace
        """
        args = copy.copy(args)
        # Get logging dir
        if not args.log_directly:
            args.logs = os.path.abspath(
                os.path.join(
                    args.logs, datetime.datetime.now().strftime("%Y_%m_%d-%H_%M_%S")
                )
            )

        # Make sure directory exists
        try:
            os.makedirs(args.logs)
        except OSError as osexc:
            if osexc.errno != errno.EEXIST:
                raise
        # Setup the basic python logging
        fprime_gds.common.logger.configure_py_log(args.logs, mirror_to_stdout=args.log_to_stdout)

        return args


class MiddleWareParser(ParserBase):
    """
    Middleware (ThreadedTcpServer) interface that looks for an address and a port. The argument handling will attempt to
    connect to the socket to ensure that it is a valid address/port and report any errors. This is then immediately
    closes the port after use. There is a minor race-condition between this check and the actual usage, however; it
    should be close enough.
    """

    @staticmethod
    def get_parser():
        """
        Sets up and parses the arguments required to run the data middleware layer. At this time, the data middleware is
        the threaded TCP server and thus the arguments are the ip address and port to listen to. May also be used in
        clients connecting to the middleware layer.

        :return: parser after arguments added
        """
        parser = argparse.ArgumentParser(
            description="Process arguments needed to specify a tool using the middleware",
            add_help=False,
        )
        # May use ZMQ transportation layer if zmq package is available
        if zmq is not None:
            parser.add_argument(
                "--zmq",
                dest="zmq",
                action="store_true",
                help="Switch to using the ZMQ transportation layer",
                default=False,
            )
            parser.add_argument(
                "--zmq-server",
                dest="zmq_server",
                action="store_true",
                help="Sets the ZMQ connection to be a server. Default: false (client)",
                default=False,
            )
            parser.add_argument(
                "--zmq-transport",
                dest="zmq_transport",
                nargs=2,
                help="Pair of URls used with --zmq to setup ZeroMQ transportation [default: %(default)s]",
                default=["ipc:///tmp/fprime-server-in", "ipc:///tmp/fprime-server-out"],
                metavar=("serverInUrl", "serverOutUrl")
            )
        parser.add_argument(
            "--tts-port",
            dest="tts_port",
            action="store",
            type=int,
            help="Set the threaded TCP socket server port [default: %(default)s]",
            default=50050,
        )
        parser.add_argument(
            "--tts-addr",
            dest="tts_addr",
            action="store",
            type=str,
            help="Set the threaded TCP socket server address [default: %(default)s]",
            default="0.0.0.0",
        )
        return parser

    @classmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Checks to ensure that the specified port and address is available before connecting. This prevents user from
        attempting to run on a port that is unavailable.

        :param args: parsed argument namespace
        :return: args namespace
        """
        # Check ZMQ settings
        if args.zmq and zmq is None:
            print("[ERROR] ZeroMQ is not available. Install pyzmq.", file=sys.stderr)
            sys.exit(-1)
        elif args.zmq:
            pass
        elif not kwargs.get("client", False):
            fprime_gds.common.communication.adapters.ip.check_port(
                args.tts_addr, args.tts_port
            )
        return args


class GdsParser(ParserBase):
    """
    Provides a parser for the following arguments:

    - dictionary: path to dictionary, either a folder for py_dicts, or a file for XML dicts
    - logs: path to logging path
    - config: configuration for GDS.

    Note: deployment can help in setting both dictionary and logs, but isn't strictly required.
    """

    @staticmethod
    def get_parser():
        """
        Creates a parser to handle the arguments required to run the GDS layer. This allows the system to start up the
        GDS and connect into the middleware layer.

        :return: parser for arguments
        """
        parser = argparse.ArgumentParser(
            description="Process arguments needed to specify a tool using the GDS",
            add_help=False,
        )
        parser.add_argument(
            "-g",
            "--gui",
            choices=GUIS,
            dest="gui",
            type=str,
            help="Set the desired GUI system for running the deployment. [default: %(default)s]",
            default="html",
        )
        parser.add_argument(
            "--dictionary",
            dest="dictionary",
            action="store",
            default=None,
            required=False,
            type=str,
            help="Path to dictionary. Overrides automatic dictionary detection.",
        )
        parser.add_argument(
            "-c",
            "--config",
            dest="config",
            action="store",
            default=None,
            type=str,
            help="Configuration for wx GUI. Ignored if not using wx.",
        )
        parser.add_argument(
            "--gui-addr",
            dest="gui_addr",
            action="store",
            default="127.0.0.1",
            required=False,
            type=str,
            help="Set the GUI server address [default: %(default)s]",
        )
        parser.add_argument(
            "--gui-port",
            dest="gui_port",
            action="store",
            default="5000",
            required=False,
            type=str,
            help="Set the GUI server address [default: %(default)s]",
        )

        return parser

    @classmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Takes the arguments from the parser, and processes them into the needed map of key to dictionaries for the
        program. This will throw if there is an error.

        :param args: parsed args into a namespace
        :return: args namespace
        """
        args = copy.copy(args)
        # Find dictionary setting via "dictionary" argument or the "deploy" argument
        if args.dictionary is not None and not os.path.exists(args.dictionary):
            raise ValueError(f"Dictionary file {args.dictionary} does not exist")

        # Handle configuration arguments
        config = fprime_gds.common.utils.config_manager.ConfigManager()
        if args.config is not None and os.path.isfile(args.config):
            config.set_configs(args.config)
        elif args.config is not None:
            raise ValueError(f"Configuration file {args.config} not found")
        args.config = config
        return args


class BinaryDeployment(ParserBase):
    """
    Parsing subclass used to read the arguments of the binary application. This derives functionality from a comm parser
    and represents the flight-side of the equation.
    """

    @staticmethod
    def get_parser():
        """
        Creates a parser to handle the arguments required to run the binary application. This allows the system to
        start up the F prime binary deployment, or ignore it.

        :return: parser for arguments
        """
        parser = argparse.ArgumentParser(
            description="Process arguments needed for running F prime binary",
            add_help=False,
        )
        parser.add_argument(
            "-n",
            "--no-app",
            dest="noapp",
            action="store_true",
            default=False,
            help="Do not run deployment binary. Overrides --app.",
        )
        parser.add_argument(
            "--app",
            dest="app",
            action="store",
            required=False,
            type=str,
            help="Path to app to run. Overrides automatic app detection.",
        )
        parser.add_argument(
            "-r",
            "--root",
            dest="root_dir",
            action="store",
            required=False,
            type=str,
            help="Root directory of build artifacts, used to automatically find app and dictionary. [default: install_dest field in settings.ini]",
        )
        return parser

    @classmethod
    def handle_arguments(cls, args, **kwargs):
        """
        Takes the arguments from the parser, and processes them into the needed map of key to dictionaries for the
        program. This will throw if there is an error.

        :param args: parsed arguments in namespace
        :return: args namespaces
        """
        # No app, stop processing now
        if args.noapp:
            return args
        args = copy.copy(args)
        if args.app is not None and not os.path.isfile(args.app):
            raise ValueError(f"F prime binary '{args.app}' does not exist")
        if args.root_dir is not None and not os.path.isdir(args.root_dir):
            raise ValueError(
                f"F prime artifacts root directory '{args.root_dir}' does not exist"
            )
        return args



"""
Created on August 16, 2019

@author: tcanham
"""

import struct
import zlib

from fprime.common.models.serialize.type_exceptions import TypeMismatchException
from fprime.common.models.serialize.numerical_types import U8Type, U16Type, U32Type
from fprime_gds.common.utils import config_manager
from fprime_gds.common.utils.data_desc_type import DataDescType


class SeqBinaryWriter:
    """
    Write out the Binary (ASTERIA) form of sequencer file.
    """

    def __init__(self, timebase=0xFFFF, config=None):
        """
        Constructor
        """
        if config is None:
            config = config_manager.ConfigManager().get_instance()

        self.__fd = None
        self.__timebase = timebase
        self.desc_obj = config.get_type("msg_desc")
        self.opcode_obj = config.get_type("op_code")
        self.len_obj = config.get_type("msg_len")

    def open(self, filename):
        """
        Open the ASCII file
        """
        self.__fd = open(filename, "wb")

    def __binaryCmdRecord(self, cmd_obj):
        """
        Return the binary command record the sequencer is expecting.
        @todo: Currently the command descriptor is always zero for immediate commands and execution time tags needed in
        command objects and seq_panel.
        """

        def __time_tag(cmd_obj):
            """
            TODO: support a timebase in the cmd obj? This is mission specific, so it is tough to handle. For now
            I am hardcoding this to 2 which is TB_NONE
            """
            # return TimeType(timeBase=2, seconds=cmd_obj.getSeconds(), useconds=cmd_obj.getUseconds()).serialize()
            # TKC - new command time format
            time = cmd_obj.get_time()
            return (
                U32Type(time.seconds).serialize()
                + U32Type(time.useconds ).serialize()
            )

        def __descriptor(cmd_obj):
            # subtract 1 from the value because enum34 enums start at 1, and this can't be changed
            return U8Type(cmd_obj.get_descriptor().value - 1).serialize()

        def __command(cmd_obj):
            self.desc_obj.val = DataDescType["FW_PACKET_COMMAND"].value
            self.opcode_obj.val = cmd_obj.get_id()
            command = (
                self.desc_obj.serialize()
            )  # serialize combuffer type enum: FW_PACKET_COMMAND
            command += self.opcode_obj.serialize()  # serialize opcode
            # Command arguments
            for arg in cmd_obj.get_args():
                command += arg.serialize()
            return command

        def __length(command):
            self.len_obj.val = len(command)
            return self.len_obj.serialize()

        def __print(byteBuffer):
            print("Byte buffer size: %d" % len(byteBuffer))
            for entry in range(0, len(byteBuffer)):
                print(
                    "Byte %d: 0x%02X (%c)"
                    % (
                        entry,
                        struct.unpack("B", byteBuffer[entry])[0],
                        struct.unpack("B", byteBuffer[entry])[0],
                    )
                )

        # This is no longer in the sequence file format.
        # def __checksum(data):
        #  csum = 0
        #  for entry in range(0,len(data)):
        #    byte = struct.unpack("B",data[entry])[0]
        #    csum += byte
        #  return U64Type(long(csum)).serialize()

        # Form header:
        descriptor = __descriptor(cmd_obj)
        time = __time_tag(cmd_obj)
        header = descriptor + time

        # Command opcode:
        command = __command(cmd_obj)

        # Command length:
        length = __length(command)

        # Checksum:
        # This is no longer in the sequence file format.
        # checksum = __checksum(header + length + command)

        # Debug printing (comment out when not debugging):
        # print "descriptor:"
        # __print(descriptor)
        # print "time:"
        # __print(time)
        # print "length:"
        # __print(length)
        # print "command:"
        # __print(command)
        # print "total record:"
        # __print(header + checksum + length + command)

        # Construct the record:
        return header + length + command

    def write(self, seq_cmds_list):
        """
        Write out each command record
        """
        num_records = len(seq_cmds_list)
        sequence = b""
        for cmd in seq_cmds_list:
            sequence += self.__binaryCmdRecord(cmd)
        size = len(sequence)
        tb_txt = b"ANY" if self.__timebase == 0xFFFF else bytes(self.__timebase)

        print("Sequence is %d bytes with timebase %s" % (size, tb_txt))

        header = b""
        header += U32Type(
            size + 4
        ).serialize()  # Write out size of the sequence file in bytes here
        header += U32Type(num_records).serialize()  # Write number of records
        header += U16Type(self.__timebase).serialize()  # Write time base
        header += U8Type(0xFF).serialize()  # write time context
        sequence = header + sequence  # Write the list of command records here
        # compute CRC. Ported from Utils/Hash/libcrc/libcrc.h (update_crc_32)
        crc = self.computeCrc(sequence)

        print("CRC: %d (0x%04X)" % (crc, crc))
        try:
            sequence += U32Type(crc).serialize()
        except TypeMismatchException as typeErr:
            print(f"Exception: {typeErr.getMsg()}")
            raise

        # Write the list of command records here
        self.__fd.write(sequence)

    def close(self):
        """
        Close the Binary file
        """
        self.__fd.close()

    def computeCrc(self, buff):
        # See http://stackoverflow.com/questions/30092226/how-to-calculate-crc32-with-python-to-match-online-results
        # RE: signed to unsigned CRC
        return zlib.crc32(buff) % (1 << 32)


class SeqAsciiWriter:
    """
    Write out the ASCII record form of sequencer file.
    """

    def __init__(self):
        """
        Constructor
        """
        self.__fd = None

    def open(self, filename):
        """
        Open the ASCII file
        """
        self.__fd = open(filename, "w")

    def __getCmdString(self, cmd_obj):
        """
        For an command return it stringified.
        """
        mnemonic = cmd_obj.getMnemonic()
        opcode = cmd_obj.getOpCode()
        args = cmd_obj.getArgs()
        #
        cmd = "{} (0x{:x})".format(mnemonic, int(opcode))
        for arg in args:
            cmd += f", {arg[2].val}"
        return cmd

    def write(self, seq_cmds_list):
        """
        Write out each record as it appears in the listbox widget.
        """
        for cmd in seq_cmds_list:
            self.__fd.write(self.__getCmdString(cmd) + "\n")

    def close(self):
        """
        Close the ASCII file
        """
        self.__fd.close()



#!/usr/bin/env python
####
# fprime_gds Python Package:
#
# The F prime GDS layer provides a basic GDS intended to enable the user to test F prime
# distributions. In addition it provides an integration and test layer to allow for automated
# testing of F prime distributions.
#
# Endpoints:
# - fprime-gds: run the F prime GDS
#
# Optional Features:
# - Test API XLS Output: Provides XLS output with the Test API. test-api-xls
# - TK GUI: if installed with Python 2, the TK GUI will be installed
#
# User Install / Upgrade:
# ```
# pip install --upgrade fprime-gds
# ```
#
# Developer and Dynamic Installation:
# ```
# pip install -e ./Gds
# ```
###

from setuptools import find_packages, setup


####
# GDS Packages:
#
# The GDS package 'tkgui' is only allowed as part of a Python 2 distribution. This code
# excludes the 'fprime_gds.tkgui' package.
####
gds_packages = find_packages("src", exclude=["*tkgui*"])
# Setup a python package using setup-tools. This is a newer (and more recommended) technology
# than distutils.
setup(
    ####
    # Package Description:
    #
    # Basic package information. Describes the package and the data contained inside. This
    # information should match the F prime description information.
    ####
    name="fprime_gds",
    use_scm_version={"root": ".", "relative_to": __file__},
    license="Apache 2.0 License",
    description="F Prime Flight Software Ground Data System layer.",
    long_description="""
This package contains the Python files used to run the F prime Ground Data System and Test API.
It is intended to supply the user with the ability to test F prime flight software in an
integrated configuration with ground in-the-loop.
    """,
    url="https://github.com/nasa/fprime",
    keywords=["fprime", "gds", "embedded", "nasa"],
    project_urls={"Issue Tracker": "https://github.com/nasa/fprime/issues"},
    # Package author, not F prime author
    author="Michael Starch",
    author_email="Michael.D.Starch@jpl.nasa.gov",
    ####
    # Included Packages:
    #
    # Will search for and included all python packages under the "src" directory.  The root package
    # is set to 'src' to avoid package names of the form src.fprime_gds. This will also ensure that
    # files included in MANIFEST.in are included in their respective packages.
    ####
    packages=gds_packages,  # See above for how GDS packages are found
    package_dir={"": "src"},
    package_data={
        "fprime_gds": ["flask/static/*", "flask/static/*/*", "flask/static/*/*/*"]
    },
    include_package_data=True,
    zip_safe=False,  # HTML templates require normal FIO access.
    ####
    # Entry Points:
    #
    # Defines the list of entry-level (scripts) that are defined by this package. This allows
    # standard use of utilities that ship as part of F prime.
    ####
    entry_points={
        "gui_scripts": ["fprime-gds = fprime_gds.executables.run_deployment:main"],
        "console_scripts": [
            "fprime-cli = fprime_gds.executables.fprime_cli:main",
            "fprime-seqgen = fprime_gds.common.tools.seqgen:main",
        ],
    },
    ####
    # Classifiers:
    #
    # Standard Python classifiers used to describe this package.
    ####
    classifiers=[
        # complete classifier list: http://pypi.python.org/pypi?%3Aaction=list_classifiers
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "Operating System :: Unix",
        "Operating System :: POSIX",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
    ],
    python_requires=">=3.7",
    setup_requires=["setuptools_scm"],
    install_requires=[
        "flask>=1.1.2, <2.0.0",
        "flask_compress>=1.11, <2.0",
        "pyzmq>=22.3.0, <23.0.0",
        "pexpect>=4.8.0, <5.0.0",
        "pytest>=6.2.4, <7.0.0",
        "flask_restful>=0.3.8, <1.0.0",
        "fprime-tools>=3.1.0a1", # Devel version until release
        "argcomplete>=1.12.3, <2.0.0",
        "Jinja2>=2.11.3, <3.0.0",
    ],
    extras_require={
        # I and T API
        "uart-adapter": "pyserial",
        "test-api-xls": "openpyxl",
    },
)



"""
@brief Encoder for cmd data

This encoder takes in cmd_data objects, serializes them, and sends the results
to all registered senders.

Serialized command format:
    +--------------------------------+          -
    | Header = "A5A5 "               |          |
    | (5 byte string)                |          |
    +--------------------------------+      Added by
    | Destination = "GUI " or "FSW " |       Sender
    | (4 byte string)                |          |
    +--------------------------------+          -
    | Command descriptor             |
    | (0x5A5A5A5A)                   |
    | (4 byte number)                |
    +--------------------------------+
    | Length of descriptor, opcode,  |
    | and argument data              |
    | (4 bytes)                      |
    +--------------------------------+
    | Descriptor type = 0            |
    | (4 bytes)                      |
    +--------------------------------+
    | Op code                        |
    | (4 bytes)                      |
    +--------------------------------+
    | Argument 1 value               |
    +--------------------------------+
    | Argument 2 value               |
    +--------------------------------+
    | ...                            |
    +--------------------------------+
    | Argument n value               |
    +--------------------------------+

@date Created July 9, 2018
@author R. Joseph Paetz

@bug No known bugs
"""


from . import encoder
from fprime.common.models.serialize.numerical_types import U32Type
from fprime_gds.common.data_types.cmd_data import CmdData
from fprime_gds.common.utils.data_desc_type import DataDescType
from fprime_gds.common.utils import config_manager


class CmdEncoder(encoder.Encoder):
    """Encoder class for command data"""

    def __init__(self, config=None):
        """
        CmdEncoder class constructor

        Args:
            config (ConfigManager, default=None): Object with configuration data
                    for the sizes of fields in the binary data. If None passed,
                    defaults are used.

        Returns:
            An initialized CmdEncoder object
        """

        if config is None:
            config = config_manager.ConfigManager().get_instance()
        super().__init__(config)

        self.len_obj = self.config.get_type("msg_len")
        self.desc_obj = self.config.get_type("msg_desc")
        self.opcode_obj = self.config.get_type("op_code")

    def encode_api(self, data):
        """
        Encodes the given CmdData object as binary data and returns the result.

        Args:
            data: CmdData object to encode

        Returns:
            Encoded version of the data argument as binary data
        """
        assert isinstance(data, CmdData), "Encoder handling incorrect type"
        cmd_temp = data.get_template()

        desc = U32Type(0x5A5A5A5A).serialize()

        self.desc_obj.val = DataDescType["FW_PACKET_COMMAND"].value
        descriptor = self.desc_obj.serialize()

        self.opcode_obj.val = cmd_temp.get_op_code()
        op_code = self.opcode_obj.serialize()

        arg_data = b""
        for arg in data.get_args():
            arg_data += arg.serialize()

        length_val = len(descriptor) + len(op_code) + len(arg_data)
        self.len_obj.val = length_val
        length = self.len_obj.serialize()

        binary_data = desc + length + descriptor + op_code + arg_data

        return binary_data



"""
dictionaries.py:

Helps the standard pipeline wrangle dictionaries by encapsulating the functionality of dictionary loading into a single
class called "Dictionaries".

@author mstarch
"""
import os

import fprime_gds.common.loaders.ch_py_loader
import fprime_gds.common.loaders.ch_xml_loader

# Py Loaders
import fprime_gds.common.loaders.cmd_py_loader

# XML Loaders
import fprime_gds.common.loaders.cmd_xml_loader
import fprime_gds.common.loaders.event_py_loader
import fprime_gds.common.loaders.event_xml_loader


class Dictionaries:
    """
    Dictionaries class to encapsulate the many different dictionaries used in the system. This includes the following
    dictionary types:

    1. Command IDs to Commands
    2. Command names to Commands
    3. Channel IDs to Channels
    4. Channel names to Channels
    5. Event IDs to Events
    6. Event names to Events
    7. Packet IDs to Packets
    """

    def __init__(self):
        """Constructor of the dictionaries object"""
        self._command_id_dict = None
        self._event_id_dict = None
        self._channel_id_dict = None
        self._command_name_dict = None
        self._event_name_dict = None
        self._channel_name_dict = None
        self._packet_dict = None

    def load_dictionaries(self, dictionary, packet_spec):
        """
        Loads the dictionaries based on the dictionary path supplied. Optional packet_spec is allowed to specify the
        definitions of packets.

        :param dictionary: dictionary path used for loading dictionaries
        :param packet_spec: specification for packets, or None, for packetized telemetry
        """
        # Loading the dictionaries from a directory. A directory indicates heritage python dicts.
        if os.path.isdir(dictionary):
            # Events
            event_loader = fprime_gds.common.loaders.event_py_loader.EventPyLoader()
            self._event_id_dict = event_loader.get_id_dict(
                os.path.join(dictionary, "events")
            )
            self._event_name_dict = event_loader.get_name_dict(
                os.path.join(dictionary, "events")
            )
            # Commands
            command_loader = fprime_gds.common.loaders.cmd_py_loader.CmdPyLoader()
            self._command_id_dict = command_loader.get_id_dict(
                os.path.join(dictionary, "commands")
            )
            self._command_name_dict = command_loader.get_name_dict(
                os.path.join(dictionary, "commands")
            )
            # Channels
            channel_loader = fprime_gds.common.loaders.ch_py_loader.ChPyLoader()
            self._channel_id_dict = channel_loader.get_id_dict(
                os.path.join(dictionary, "channels")
            )
            self._channel_name_dict = channel_loader.get_name_dict(
                os.path.join(dictionary, "channels")
            )
        # XML dictionaries
        elif os.path.isfile(dictionary):
            # Events
            event_loader = fprime_gds.common.loaders.event_xml_loader.EventXmlLoader()
            self._event_id_dict = event_loader.get_id_dict(dictionary)
            self._event_name_dict = event_loader.get_name_dict(dictionary)
            # Commands
            command_loader = fprime_gds.common.loaders.cmd_xml_loader.CmdXmlLoader()
            self._command_id_dict = command_loader.get_id_dict(dictionary)
            self._command_name_dict = command_loader.get_name_dict(dictionary)
            # Channels
            channel_loader = fprime_gds.common.loaders.ch_xml_loader.ChXmlLoader()
            self._channel_id_dict = channel_loader.get_id_dict(dictionary)
            self._channel_name_dict = channel_loader.get_name_dict(dictionary)
        else:
            raise Exception(f"[ERROR] Dictionary '{dictionary}' does not exist.")
        # Check for packet specification
        if packet_spec is not None:
            packet_loader = fprime_gds.common.loaders.pkt_xml_loader.PktXmlLoader()
            self._packet_dict = packet_loader.get_id_dict(
                packet_spec, self._channel_name_dict
            )
        else:
            self._packet_dict = None

    @property
    def command_id(self):
        """Command dictionary by ID"""
        return self._command_id_dict

    @property
    def event_id(self):
        """Event dictionary by ID"""
        return self._event_id_dict

    @property
    def channel_id(self):
        """Channel dictionary by ID"""
        return self._channel_id_dict

    @property
    def command_name(self):
        """Command dictionary by name"""
        return self._command_name_dict

    @property
    def event_name(self):
        """Event dictionary by name"""
        return self._event_name_dict

    @property
    def channel_name(self):
        """Channel dictionary by name"""
        return self._channel_name_dict

    @property
    def packet(self):
        """Packet dictionary"""
        return self._packet_dict



"""
@brief Class to parse raw messages and distribute them to registered objects

This class allows for other objects (namely Decoders) to register to receive
data messages of a given descriptor type. The data message with the length and
descriptor header will be passed on to the registered objects.

@date Created June 2018
@date Modified August 2, 2018

@author Josef Biberstein
@author Joseph Paetz

@bug No known bugs
"""
import logging
from fprime_gds.common.utils import config_manager, data_desc_type
from fprime_gds.common.handlers import DataHandler
from fprime_gds.common.decoders.decoder import DecodingException


LOGGER = logging.getLogger("distributor")


# NOTE decoder function to call is called data_callback(data)
class Distributor(DataHandler):
    """
    A distributor contains a socket client that connects to a ThreadedTCPServer.
    It then sends and recvs data from a FPrime deployment.
    Decoders can register with a distributor to recv packets of data of a certain description.
    """

    def __init__(self, config=None):
        """
        Sets up the dictionary of connected decoders and socket client object.

        Decoder dictionary is of the form:
        {data descriptor name: list of decoder objects registered for that data}

        Args:
            config (ConfigManager, default=None): Config manager with
                   information on what types the message fields are. If None,
                   defaults are used.
        """
        if config is None:
            # Retrieve singleton for the configs, or defaults if singleton unused
            config = config_manager.ConfigManager().get_instance()

        self.__decoders = {key.name: [] for key in list(data_desc_type.DataDescType)}

        # Internal buffer for un distributed data
        self.__buf = bytearray(b"")
        # Setup key framing
        self.key_frame = None
        tmp_frame = config.get("framing", "use_key", fallback="false")
        if tmp_frame.lower() == "true":
            self.key_frame = int(config.get("framing", "key_val"), 16)
        self.key_obj = config.get_type("key_val")
        self.len_obj = config.get_type("msg_len")
        self.desc_obj = config.get_type("msg_desc")

    # NOTE we could use either the type of the object or an enum as the type argument.
    # It should indicate what the decoder decodes.

    def register(self, typeof, obj):
        """
        Register a decoder with the distributor

        Arguments:
            typeof {string} -- The name of the data descriptor that the decoder will decode
            obj {decoder} -- The decoder object that will process the data
        """
        self.__decoders[typeof].append(obj)

    def parse_into_raw_msgs_api(self, data):
        """
        Parse the given data into raw messages.

        Raw messages include a length and descriptor header in front of the
        message data.

        Args:
            data (bytearray): Binary data to parse. First byte of data should be
                              the first byte of a valid raw message.

        Returns:
            (leftover_data, [raw_msg1, raw_msg2, ..., raw_msgN])
            Where leftover_data is a bytearray of anything at the end of data
            that could not be parsed (due to insufficient length).
        """
        data_left = data

        raw_msgs = []
        # Keep parsing and then break when you can't parse no more
        while True:
            # Search data looking for key-frame
            if self.key_frame is not None:
                while True:
                    # Check if we have enough data to parse a key
                    # if not, bail on the function
                    if len(data_left) < self.key_obj.getSize():
                        return data_left, raw_msgs
                    # Check leading key size bytes to see if it is the key
                    self.key_obj.deserialize(data_left, 0)
                    if self.key_obj.val != self.key_frame:
                        del data_left[0]
                        continue
                    # Key found break
                    del data_left[: self.key_obj.getSize()]
                    break

            # Check if we have enough data to parse a length
            if len(data_left) < self.len_obj.getSize():
                break
            self.len_obj.deserialize(data_left, 0)
            expected_len = self.len_obj.val + self.len_obj.getSize()

            # Check if we have enough data to parse
            if len(data_left) < expected_len:
                break

            raw_msgs.append(data_left[:expected_len])
            del data_left[:expected_len]
        return data_left, raw_msgs

    def parse_raw_msg_api(self, raw_msg):
        """
        Parse the given raw message into its component parts

        Args:
            raw_msg (bytearray): Raw message to parse. Must be a valid raw msg

        Returns:
            Tuple: (length, descriptor, msg). length is type int. Descriptor is
            type int. Msg is a bytearray.
        """
        # Data Entry Structure
        #
        # +---------------------------+
        # | Length (n bytes)          |
        # +---------------------------+
        # | Descriptor Type (4 bytes) |      -
        # +---------------------------+      |
        # |                           |      |
        # | Message data...           |   Length
        # | ...                       |      |
        # | ..                        |      |
        #   .                                :

        offset = 0

        # Parse length
        self.len_obj.deserialize(raw_msg, offset)
        offset += self.len_obj.getSize()
        length = self.len_obj.val

        # Parse Descriptor type
        self.desc_obj.deserialize(raw_msg, offset)
        offset += self.desc_obj.getSize()
        desc = self.desc_obj.val

        # Retrieve message section
        msg = raw_msg[offset:]

        return length, desc, msg

    def data_callback(self, data, sender=None):
        """Delegates to on_recv after checking data is binary"""
        assert isinstance(
            data, (bytes, bytearray)
        ), "Distributor must be registered to handle binary data only"
        return self.on_recv(data)

    def on_recv(self, data):
        """
        Called by the internal socket client when data is received from the socket
        client.

        Arguments:
            data {binary} -- the data received from the socket client. May contain
                             more than one message.
        """
        # NOTE make data sizes selectable with a configuration later
        # NOTE: Currently, the TCPServer sends the client just the raw bytes,
        #       without headers. This means that we don't have a good way to
        #       figure out where the head of the messages are when we decode
        #       them here. Ideally, the client would just be sending individual
        #       messages. This is handled in the comm layer, sending individual
        #       messages through the TCP Server.

        # Add new data to end of buffer

        self.__buf.extend(data)

        (leftover_data, raw_msgs) = self.parse_into_raw_msgs_api(self.__buf)
        assert leftover_data == self.__buf, "Leftover data is not equivalent to the remaining data in buffer"

        for raw_msg in raw_msgs:
            (length, data_desc, msg) = self.parse_raw_msg_api(raw_msg)

            data_desc_key = data_desc_type.DataDescType(data_desc).name

            for d in self.__decoders[data_desc_key]:
                try:
                    d.data_callback(msg)
                except DecodingException as dexc:
                    LOGGER.warning("Decoding error occurred: %s. Skipping.", dexc)


####
# events.py:
#
# This file captures the HTML endpoint for the events list. This file contains one endpoint which allows for listing
# events after a certain time-stamp.
#
#  GET /events: list events
#      Input Data: {
#                      "start-time": "YYYY-MM-DDTHH:MM:SS.sss" #Start time for event listing
#                  }
####
import copy

from fprime_gds.flask.resource import DictionaryResource, HistoryResourceBase
from fprime_gds.common.utils.string_util import format_string_template


class EventDictionary(DictionaryResource):
    """Channel dictionary shares implementation"""


class EventHistory(HistoryResourceBase):
    """
    Resource supplying the history of events in the system. Includes `get_display_text` postprocessing to add in the
    getter for the display text.
    """

    def process(self, event):
        """Process item and return one with get_display_text"""
        event = copy.copy(event)
        setattr(
            event,
            "display_text",
            format_string_template(
                event.template.format_str, tuple([arg.val for arg in event.args])
            ),
        )
        return event



import os
import sys
import time
import unittest

from fprime.common.models.serialize.numerical_types import I32Type
from fprime.common.models.serialize.time_type import TimeType
from fprime_gds.common.data_types.ch_data import ChData
from fprime_gds.common.history.chrono import ChronologicalHistory
from fprime_gds.common.templates.ch_template import ChTemplate
from fprime_gds.common.testing_fw import predicates

filename = os.path.dirname(__file__)
gdsName = os.path.join(filename, "../../../../src")
fprimeName = os.path.join(filename, "../../../../../Fw/Python/src")
sys.path.insert(0, gdsName)
sys.path.insert(0, fprimeName)


"""
These unit tests were quickly modified from the original tests on the predicate history.
They all run, but may not be written in an optimized way.
"""


class HistoryTestCases(unittest.TestCase):
    def setUp(self):
        self.cHistory = ChronologicalHistory()

    @staticmethod
    def get_range(length):
        temp1 = ChTemplate(1, "Test Channel 1", "Chrono_Hist_Tester", I32Type)
        chList = []
        ts0 = TimeType()
        for item in range(length):
            tsi = ts0 + time.time()
            chList.append(ChData(I32Type(item), tsi, temp1))
        return chList

    @staticmethod
    def assert_lists_equal(expected, actual):
        assert len(expected) == len(
            actual
        ), f"the given list should have had the length {len(expected)}, but instead had {len(actual)}"
        for i in range(len(expected)):
            assert (
                expected[i] is actual[i]
            ), f"the {i} element of the expected list should be {expected[i]}, but was {actual[i]}."

    def test_push_and_retrieve(self):
        self.assert_lists_equal([], self.cHistory.retrieve())
        tList = []
        for item in self.get_range(100):
            self.cHistory.data_callback(item)
            tList.append(item)
        self.assert_lists_equal(tList, self.cHistory.retrieve())

        self.cHistory.clear()
        rList = list(reversed(tList))
        for item in rList:
            self.cHistory.data_callback(item)
        self.assert_lists_equal(tList, self.cHistory.retrieve())

        self.cHistory.clear()
        self.assert_lists_equal([], self.cHistory.retrieve())

    def test_push_and_retrieve_at_index(self):
        tList = []
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)
        self.assert_lists_equal(tList[10:], self.cHistory.retrieve(10))

        self.cHistory.clear()
        rList = list(reversed(tList))
        for item in rList:
            self.cHistory.data_callback(item)
        self.assert_lists_equal(tList[10:], self.cHistory.retrieve(10))

    def test_push_and_retrieve_from_predicate(self):
        tList = []
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)
        start = predicates.equal_to(tList[45])
        self.assert_lists_equal(tList[45:], self.cHistory.retrieve(start))

    def test_push_and_clear(self):
        tList = []
        for item in self.get_range(200):
            self.cHistory.data_callback(item)
            tList.append(item)
        self.assert_lists_equal(tList, self.cHistory.retrieve())
        self.cHistory.clear()
        self.assert_lists_equal([], self.cHistory.retrieve())

    def test_push_and_clear_index(self):
        tList = []
        for item in self.get_range(200):
            self.cHistory.data_callback(item)
            tList.append(item)
        self.assert_lists_equal(tList, self.cHistory.retrieve())
        self.cHistory.clear(100)
        self.assert_lists_equal(tList[100:], self.cHistory.retrieve())

    def test_push_and_clear_predicate(self):
        tList = []
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)
        self.assert_lists_equal(tList, self.cHistory.retrieve())
        start = predicates.equal_to(tList[45])
        self.cHistory.clear(start)
        self.assert_lists_equal(tList[45:], self.cHistory.retrieve())

    def test_history_size(self):
        assert self.cHistory.size() == 0, "starting history is empty"
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
        assert self.cHistory.size() == 50, "starting history is empty"
        self.cHistory.clear(25)
        assert self.cHistory.size() == 25, "starting history is empty"

    def test_history_retrieve_new(self):
        tList = []
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)

        self.assert_lists_equal(tList, self.cHistory.retrieve())
        tList.clear()

        for item in self.get_range(30):
            self.cHistory.data_callback(item)
            tList.append(item)

        self.assert_lists_equal(tList, self.cHistory.retrieve_new())
        tList.clear()

        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)

        self.assert_lists_equal(tList, self.cHistory.retrieve_new())
        self.assert_lists_equal([], self.cHistory.retrieve_new())

    def test_history_retrieve_new_repeat(self):
        tList = self.get_range(50)
        cHistory2 = ChronologicalHistory()
        for item in tList[1:]:
            self.cHistory.data_callback(item)
            cHistory2.data_callback(item)

        self.assert_lists_equal(tList[1:], self.cHistory.retrieve_new(True))
        self.assert_lists_equal(tList[1:], cHistory2.retrieve_new(False))

        self.cHistory.data_callback(tList[0])
        cHistory2.data_callback(tList[0])

        self.assert_lists_equal(tList, self.cHistory.retrieve_new(True))
        self.assert_lists_equal([tList[0]], cHistory2.retrieve_new(False))

    def test_history_retrieve_new_after_clear(self):
        tList = []
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)

        self.assert_lists_equal(tList, self.cHistory.retrieve())
        tList.clear()

        for item in self.get_range(50):
            self.cHistory.data_callback(item)
            tList.append(item)

        self.cHistory.clear(75)

        self.assert_lists_equal(tList[25:], self.cHistory.retrieve_new())
        tList.clear()

    def test_history_iterable(self):
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
        tList = []
        for item in self.cHistory:
            tList.append(item)
        self.assert_lists_equal(tList, self.cHistory.retrieve_new())

    def test_history_length(self):
        assert len(self.cHistory) == 0, "starting history is empty"
        for item in self.get_range(50):
            self.cHistory.data_callback(item)
        assert len(self.cHistory) == 50, "starting history is empty"
        self.cHistory.clear(25)
        assert len(self.cHistory) == 25, "starting history is empty"

    def test_history_filter(self):
        class is_even(predicates.predicate):
            def __call__(self, item):
                return item.get_val() % 2 == 0

            def __str__(self):
                return "Decides if a value is even"

        pred = is_even()
        self.cHistory = ChronologicalHistory(pred)
        for item in self.get_range(50):
            self.cHistory.data_callback(item)

        assert len(self.cHistory) == 25

        correct_error = False
        try:
            ChronologicalHistory("Not a predicate")
        except TypeError:
            correct_error = True
        assert correct_error, "The History should have raised a TypeError"


if __name__ == "__main__":
    unittest.main()



"""
@brief Base class for system data classes.

This class defines the interface for cata classes which are intended to hold
a specific data item (packet, channel, event). This data item includes the time
of the data as well as data such as channel value or argument value.

@date Created July 2, 2018
@author R. Joseph Paetz (rpaetz@jpl.nasa.gov)

@bug No known bugs
"""
from fprime.common.models.serialize import time_type
from fprime_gds.common.templates import data_template


class SysData:
    """
    The SysData class defines the interface for system data classes which are
    for specific data readings/events
    """

    def __init__(self):
        """
        Constructor.

        Each subclass will define new constructors with necessary arguments.
        The necessary fields are time, id, and template.

        Returns:
            An initialized SysData object
        """
        if not hasattr(self, "id"):
            self.id = 0
        if not hasattr(self, "template"):
            self.template = data_template.DataTemplate()
        if not hasattr(self, "time"):
            self.time = time_type.TimeType()

    def get_id(self):
        """
        Returns the id of the channel

        Returns:
            The id of the channel
        """
        return self.id

    def get_time(self):
        """
        Returns the time of the channel data reading

        Returns:
            Time of the reading as a TimeType
        """
        return self.time

    def get_template(self):
        """
        Returns the template class instance for the data stored

        Returns:
            An instance of a template class for this instance's data
        """
        return self.template

    @staticmethod
    def compare(x, y):
        """
        Compares two data items.

        Returns:
            Negative, 0, or positive for t1<t2, t1==t2, t1>t2 respectively
        """
        # Compare by time first
        time_comp = time_type.TimeType.compare(x.time, y.time)

        if time_comp != 0:
            return time_comp

        # Compare by id second (just let multiple events at the same time with
        # the same id be counted as equal
        return time_type.TimeType.compare(x.id, y.id)



"""
__init__.py:

Python Logging Setup. This sets up the global logging format for all Gds components. This allows the user to use Python
loggers without calling basic config.

@author mstarch
"""
import logging
import os
import sys


def configure_py_log(directory=None, filename=sys.argv[0], mirror_to_stdout=False):
    """
    Configure the python logging. If logdir is supplied, our logs will go in that directory as a log file. Otherwise,
    logs will go to the CLI.

    :param directory: directory logs are written into
    :param filename: logging filename
    :param mode: of file to write
    :param mirror_to_stdout: mirror the log output to standard our
    """
    handlers = [logging.StreamHandler(sys.stdout)] if directory is None or mirror_to_stdout else []
    if directory is not None:
        log_file = os.path.join(directory, os.path.basename(filename))
        log_file = log_file if log_file.endswith(".log") else f"{log_file}.log"
        handlers.append(logging.FileHandler(log_file))
    formatter = logging.Formatter("[%(asctime)s] [%(levelname)s] %(name)s: %(message)s")

    for handler in handlers:
        handler.setFormatter(formatter)
        logging.getLogger().addHandler(handler)
    logging.getLogger().setLevel(logging.INFO)
    logging.info("Logging system initialized!")




"""
Created on Jan 9, 2015
@author: reder
"""
# Exception classes for all controllers modules


class GseControllerException(Exception):
    def __init__(self, val):
        self.except_msg = val
        super().__init__(val)

    def getMsg(self):
        return self.except_msg


class GseControllerUndefinedDirectoryException(GseControllerException):
    def __init__(self, val):
        super().__init__(f"Path does not exist: {str(val)}!")


class GseControllerUndefinedFileException(GseControllerException):
    def __init__(self, val):
        super().__init__(f"Path does not exist: {str(val)}!")


class GseControllerParsingException(GseControllerException):
    def __init__(self, val):
        super().__init__(f"Parsing error: {str(val)}")


class GseControllerMnemonicMismatchException(GseControllerException):
    def __init__(self, val1, val2):
        super().__init__(f"ID mismatch ({str(val1)}, {str(val2)})!")


class GseControllerStatusUpdateException(GseControllerException):
    def __init__(self, val):
        super().__init__(f"Bad status update mode: {str(val)}!")



"""
@brief Loader class for importing xml based channel dictionaries

@date Created July 19, 2018
@author R. Joseph Paetz

@bug No known bugs
"""

from fprime_gds.common.data_types import exceptions
from fprime_gds.common.templates.ch_template import ChTemplate

# Custom Python Modules
from .xml_loader import XmlLoader


class ChXmlLoader(XmlLoader):
    """Class to load xml based telemetry channel dictionaries"""

    CH_SECT = "channels"  # Name of the channel section

    # Xml Tags for important fields
    COMP_TAG = "component"
    NAME_TAG = "name"
    ID_TAG = "id"
    DESC_TAG = "description"
    TYPE_TAG = "type"
    FMT_STR_TAG = "format_string"
    LOW_R_TAG = "low_red"
    LOW_O_TAG = "low_orange"
    LOW_Y_TAG = "low_yellow"
    HIGH_Y_TAG = "high_yellow"
    HIGH_O_TAG = "high_orange"
    HIGH_R_TAG = "high_red"

    def construct_dicts(self, path):
        """
        Constructs and returns python dictionaries keyed on id and name

        This function should not be called directly, instead, use
        get_id_dict(path) and get_name_dict(path)

        Args:
            path: Path to the xml dictionary file containing channel information

        Returns:
            A tuple with two channel dictionaries (python type dict):
            (id_idct, name_dict). The keys are the channels' id and name fields
            respectively and the values are ChTemplate objects
        """
        xml_tree = self.get_xml_tree(path)

        # Check if xml dict has channels section
        ch_section = self.get_xml_section(self.CH_SECT, xml_tree)
        if ch_section is None:
            raise exceptions.GseControllerParsingException(
                f"Xml dict did not have a {self.CH_SECT} section"
            )

        id_dict = {}
        name_dict = {}
        for ch in ch_section:
            ch_dict = ch.attrib

            # Assume the required fields are present (component, name, id,
            #  description, type). Check for all others
            ch_comp = ch_dict[self.COMP_TAG]
            ch_name = ch_dict[self.NAME_TAG]
            ch_id = int(ch_dict[self.ID_TAG], base=16)
            ch_type_obj = self.parse_type(ch_dict[self.TYPE_TAG], ch, xml_tree)

            ch_desc = None
            ch_fmt_str = None
            ch_low_red = None
            ch_low_orange = None
            ch_low_yellow = None
            ch_high_yellow = None
            ch_high_orange = None
            ch_high_red = None

            if self.DESC_TAG in ch_dict:
                ch_desc = ch_dict[self.DESC_TAG]

            if self.FMT_STR_TAG in ch_dict:
                ch_fmt_str = ch_dict[self.FMT_STR_TAG]

            # TODO we need to convert these into numbers, is this the best
            #  way to do it?
            if self.LOW_R_TAG in ch_dict:
                ch_low_red = float(ch_dict[self.LOW_R_TAG])

            if self.LOW_O_TAG in ch_dict:
                ch_low_orange = float(ch_dict[self.LOW_O_TAG])

            if self.LOW_Y_TAG in ch_dict:
                ch_low_yellow = float(ch_dict[self.LOW_Y_TAG])

            if self.HIGH_Y_TAG in ch_dict:
                ch_high_yellow = float(ch_dict[self.HIGH_Y_TAG])

            if self.HIGH_O_TAG in ch_dict:
                ch_high_orange = float(ch_dict[self.HIGH_O_TAG])

            if self.HIGH_R_TAG in ch_dict:
                ch_high_red = float(ch_dict[self.HIGH_R_TAG])

            ch_temp = ChTemplate(
                ch_id,
                ch_name,
                ch_comp,
                ch_type_obj,
                ch_fmt_str,
                ch_desc,
                ch_low_red,
                ch_low_orange,
                ch_low_yellow,
                ch_high_yellow,
                ch_high_orange,
                ch_high_red,
            )

            id_dict[ch_id] = ch_temp
            name_dict[ch_temp.get_full_name()] = ch_temp

        return id_dict, name_dict



"""
helpers.py:

Helpers for file uplink, and downlink. This contains shared components, classes, and support architecture in order to
enable both file uplink and downlink to share the same structures. This includes the following shared objects:

1. Timeout: for managing timeouts and timeout responses
2. FileStates: managing the state of an uplink or downlinking file
3. CFDPChecksum: calculates the CFDP checksum for files
4. TransmitFile:  file object for up and down

@author mstarch, and Blake A. Harriman's work
"""
import datetime
import enum
import logging
import os
import struct
import threading
from pathlib import Path


class Timeout:
    """
    Starts a timeout thread and will respond with a callback to a function when the timeout expires.
    """

    def __init__(self):
        """Sets up needed member variables"""
        self.__timer = None
        self.__timeout = None
        self.__callback = None
        self.args = ()

    def setup(self, callback, timeout=5, args=()):
        """
        Sets up the timeout but does not start it.

        :param callback: function called when timeout expires
        :param timeout: (optional) timeout duration. Default: 5 seconds
        :param args: args supplied to the python timer object
        """
        self.__timeout = timeout
        self.__callback = callback
        self.args = args

    def start(self):
        """Starts the timeout after a previous setup."""
        assert self.__timer is None, "Timer already started, call restart() instead"
        assert self.__callback is not None, "Setup timeout before calling start"
        assert self.__timeout is not None, "Setup timeout before calling start"
        self.__timer = threading.Timer(self.__timeout, self.__callback, args=self.args)
        self.__timer.start()

    def restart(self):
        """
        Restarts the given timer first canceling and then starting it again.
        """
        self.stop()
        self.start()

    def stop(self):
        """
        Stops the timeout preventing a callback to the stored function.
        """
        if self.__timer is not None:
            self.__timer.cancel()
            self.__timer = None


class FileStates(enum.Enum):
    """
    An enumeration of states used in the file uplinker and downlinker.
    """

    IDLE = 0
    RUNNING = 1
    CANCELED = 2
    END_WAIT = 3  # Waiting for the handshake for CANCEL or END packet


class CFDPChecksum:
    """Class running the CFDG checksum"""

    def __init__(self):
        """Set initial value as zero"""
        self.__value = 0

    def update(self, data, offset):
        """Update the checksum"""
        while data:
            padding_len = offset % 4
            calc_bytes = (
                bytes([0] * padding_len) + data[: 4 - padding_len] + bytes([0, 0, 0, 0])
            )
            self.__value = (
                self.__value + struct.unpack_from(">I", calc_bytes, 0)[0]
            ) & 0xFFFFFFFF
            # Update pointers
            data = data[4 - padding_len :]
            offset = offset + (4 - padding_len)

    @property
    def value(self):
        return self.__value


class TransmitFileState(enum.Enum):
    READ = 0
    WRITE = 1


class TransmitFile:
    """
    Wraps the file information needed for the uplink and downlinking processes.
    """

    def __init__(self, source, destination, size=None, log_dir=None):
        """Construct the uplink file"""
        self.__mode = None
        self.__start = None
        self.__end = None
        self.__source = source
        self.__destination = destination
        self.__size = size if size is not None else os.path.getsize(source)
        self.__seek = 0
        self.__state = "QUEUED"
        self.__fd = None
        self.__checksum = CFDPChecksum()
        self.__log_dir = log_dir
        self.__log_handler = None

    def open(self, mode):
        """
        Opens the file descriptor and prepares it for uplink/downlink
        """
        assert self.__fd is None, "Must close file before attempting to reopen it"
        self.__mode = mode
        if mode == TransmitFileState.WRITE:
            filepath = self.__destination
            Path(filepath).touch(exist_ok=True)
            filemode = "rb+"
        else:
            filepath = self.__source
            filemode = "rb"

        self.__state = "TRANSMITTING"
        self.__fd = open(filepath, filemode)
        self.__start = datetime.datetime.utcnow()
        if self.__log_dir is not None:
            self.__log_handler = logging.FileHandler(
                os.path.join(self.__log_dir, f"{os.path.basename(filepath)}.log"),
                "a",
            )

    def read(self, chunk):
        """Read the chunk from the file"""
        assert self.__fd is not None, "Must open file before reading"
        assert self.__mode == TransmitFileState.READ, "Cannot read in WRITE mode"
        return self.__fd.read(chunk)

    def write(self, chunk, offset):
        """
        Write a chunk to the file.

        :param chunk: data to write to the file
        :param offset: offset to write to
        """
        assert self.__fd is not None, "Must open file before writing"
        assert self.__mode == TransmitFileState.WRITE, "Cannot write in READ mode"
        self.__fd.seek(offset)
        self.__fd.write(chunk)

    def close(self):
        """
        Opens the file descriptor and prepares it for uplink
        """
        if self.__log_handler is not None:
            self.__log_handler.close()
            self.__log_handler = None
        if self.__fd is not None:
            self.__fd.close()
            self.__fd = None
            self.__end = datetime.datetime.utcnow()

    @property
    def start(self):
        return self.__start

    @property
    def end(self):
        return self.__end

    @property
    def source(self):
        return self.__source

    @property
    def destination(self):
        return self.__destination

    @property
    def size(self):
        return self.__size

    @property
    def seek(self):
        return self.__seek

    @seek.setter
    def seek(self, seek):
        self.__seek = seek

    @property
    def state(self):
        return self.__state

    @state.setter
    def state(self, state):
        self.__state = state

    @property
    def checksum(self):
        return self.__checksum

    @property
    def log_handler(self):
        return self.__log_handler


def file_to_dict(files, uplink=True):
    """
    Converts files to dictionary. This creates a new list of JSONable file dictionaries.

    :param files: list of TransmitFiles to convert
    :param uplink: is the file being uplinked. Default: True
    :return: list of dictionaries
    """
    return [
        {
            "source": item.source,
            "destination": item.destination,
            "size": item.size,
            "current": item.seek,
            "state": item.state,
            "percent": 100
            if item.size == 0
            else int(item.seek / item.size * 100.0),
            "uplink": uplink,
            "start": item.start,
            "end": item.end,
        }
        for item in files
    ]
